package biz.hammurapi.dataflow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;

/**
 * Base class for mixers which join data.
 * @author Pavel
 */
public abstract class AbstractMixer implements DataSource {
	
	/**
	 * Joins inputs.
	 * @author Pavel
	 *
	 */
	public interface JoinHelper {
		
		void join(InputEntry[] inputs, String activatorName, DataSink sink);

	}
	
	/**
	 * Decides whether given inputs shall be joined.
	 * @author Pavel
	 */
	public interface JoinPredicate {
		
		boolean toBeJoined(InputEntry[] inputs, String activatorName);

	}
	
	protected DataSink sink;
	
	public void setSink(DataSink sink) {
		this.sink = sink;		
	}

	// Keeps inputs preserving addition order.
	private LinkedHashMap inputs = new LinkedHashMap();
	
	private class MixerInput implements DataSink {
		
		String name;		
		
		Collection data;
		
		public MixerInput(String name) {
			this.name = name;
			data = createDataCollection(name);
		}

		public boolean addData(Data data) {
			return AbstractMixer.this.addData(name, data);
		}

		public boolean controlsMultiplexing() {
			return AbstractMixer.this.controlsMultiplexing(name);
		}

		public DataItemInfo[] getDataInfo() {
			return AbstractMixer.this.getDataInfo(name);
		}
		
	}
	
	/**
	 * Information about input presented to join().
	 * @author Pavel
	 *
	 */
	protected interface InputEntry {
		
		/**
		 * @return Input name.
		 */
		String getName();
		
		/**
		 * @return Input data.
		 */
		Data getData();
		
		/**
		 * Join handler invokes this method to indicate that given
		 * data item shall not be used in any further joins, i.e. it's been
		 * "consumed".
		 */
		void removeFromJoin();
		
	}
	
	/**
	 * Abstract mixer presents all permutations of newly arrived data from given input 
	 * with previously arrived data from other inputs plus null values to include cases
	 * with inputs without data.  
	 * @param inputs Inputs for joining
	 * @param activatorName name of the input which activated this join (newly arrived data)
	 * @param sink DataSink to write results of joining to (if any).
	 */
	protected abstract void join(InputEntry[] inputs, String activatorName, DataSink sink);
		
	/**
	 * @param name
	 * @return Named input. Inputs are created on request.
	 */
	public DataSink getInput(String name) {
		DataSink ret = (DataSink) inputs.get(name);
		if (ret==null) {
			ret = new MixerInput(name);
			inputs.put(name, ret);
		}
		
		return ret;
	}
	
	/**
	 * @param name Input name.
	 * @return multiplexing control flag for given named input.
	 */
	protected boolean controlsMultiplexing(String name) {
		return false;
	}
	
	/**
	 * @param name Input name.
	 * @return Metadata for given named input.
	 */
	protected DataItemInfo[] getDataInfo(String name) {
		return null;
	}
	
	/**
	 * Creates collection to hold input data for joining.
	 * This implementation creates ArrayList. Override to create other types
	 * of collections, e.g. Set or persistent collection for long-running, restartable
	 * data flows.
	 * @param name Input name.
	 * @return collection to hold input data for joining.
	 */
	protected Collection createDataCollection(String name) {
		return new ArrayList();
	}

	/**
	 * This method invokes join logic.
	 * @param name Input name.
	 * @param data Input data.
	 * @return Always returns false - never consumes input. Override if needed.
	 */
	protected boolean addData(String name, Data data) {
		InputEntry[] inputEntries = new InputEntry[this.inputs.size()];
		MixerInput[] mixerInputs = new MixerInput[inputEntries.length];
		Iterator it = inputs.values().iterator();
		for (int i=0; it.hasNext(); ++i) {
			mixerInputs[i] = (MixerInput) it.next();
		}
		
		// Start joining data.
		doJoin(name, data, inputEntries, mixerInputs, 0);
		
		return false;
	}
	
	/**
	 * Iterates over all data items to pass eventually to join().
	 * Method is synchronized to avoid concurrent modification exceptions.
	 * @param name Activator item name.
	 * @param data Activator data.
	 * @param inputEntries Input entries
	 * @param mixerInputs Mixer inputs.
	 * @param inputIdx Index of currently processed input.
	 * @return If one of entries was returned from join the method returns index
	 * of that entry. All entries with higher index shall stop processing further
	 * join.
	 */
	private synchronized int doJoin(
			String name, 
			Data data, 
			InputEntry[] inputEntries,
			MixerInput[] mixerInputs, 
			int inputIdx) {
				
		class InputEntryImpl implements InputEntry {
			
			String name;
			Data data;
			boolean removedFromJoin;
			
			InputEntryImpl(String name, Data data) {
				this.data = data;
				this.name = name;
			}
			
			void setData(Data data) {
				this.data = data;
			}

			boolean isRemovedFromJoin() {
				return removedFromJoin;
			}
			
			public Data getData() {
				return data;
			}

			public String getName() {
				return name;
			}

			public void removeFromJoin() {
				removedFromJoin = true;				
			}

			void resetRemoveFromJoin() {
				removedFromJoin = false;			
			}
			
		}
		
		if (inputIdx == inputEntries.length) {
			join(inputEntries, name, sink);
			for (int i=0; i<inputEntries.length; ++i) {
				if (inputEntries[i] instanceof InputEntryImpl && ((InputEntryImpl) inputEntries[i]).isRemovedFromJoin()) {
					return i;
				}
			}
		} else {
			if (mixerInputs[inputIdx].name.equals(name)) {
				InputEntryImpl currentEntry = new InputEntryImpl(name, data);
				inputEntries[inputIdx] = currentEntry;
				int result = doJoin(name, data, inputEntries, mixerInputs, inputIdx+1);
				if (!currentEntry.isRemovedFromJoin()) {
					mixerInputs[inputIdx].data.add(data);
				} 
				return result;
			} else {
				inputEntries[inputIdx] = null;
				int result = doJoin(name, data, inputEntries, mixerInputs, inputIdx+1); // Join with null first.
				if (result<inputIdx) {
					return result;
				}
				InputEntryImpl currentEntry = new InputEntryImpl(name, null);
				inputEntries[inputIdx] = currentEntry;
				Iterator it = mixerInputs[inputIdx].data.iterator();
				while (it.hasNext()) {
					currentEntry.setData((Data) it.next());
					result = doJoin(name, data, inputEntries, mixerInputs, inputIdx+1);
					if (currentEntry.isRemovedFromJoin()) {
						it.remove();
						currentEntry.resetRemoveFromJoin();
					}
					if (result<inputIdx) {
						return result;
					}
				}
			}
		}
		return inputEntries.length;
	}

	/**
	 * Factory method for data. This implementation creates MapData. Override
	 * to create different type of Data, e.g. persistent data backed by relational
	 * database.
	 * @param chain Chain of data to delegate data item requests. Can be null. 
	 * @return
	 */
	protected Data createData(Data[] chain) {
		return chain==null ? new MapData() : new MapData(chain);
	}
	
	/**
	 * Creates output data by mounting non-null inputs using their names plus separator
	 * as prefixes.
	 * @param inputs
	 * @return
	 */
	protected Data mountInputs(InputEntry[] inputs, String separator) {
		Data ret = createData(null);
		for (int i=0; i<inputs.length; ++i) {
			if (inputs[i]!=null) {
				Data iData = inputs[i].getData();
				if (iData!=null) {
					ret.mount(inputs[i].getName()+separator, iData);
				}
			}
		}
		return ret;
	}

	/**
	 * Creates output data by chaining non-null inputs.
	 * @param inputs
	 * @return
	 */
	protected Data chainInputs(InputEntry[] inputs) {
		List chain = new ArrayList();
		for (int i=0; i<inputs.length; ++i) {
			if (inputs[i]!=null) {
				Data iData = inputs[i].getData();
				if (iData!=null) {
					chain.add(iData);
				}
			}
		}
		if (chain.isEmpty()) {
			return createData(null);
		}
		return createData((Data[]) chain.toArray(new Data[chain.size()]));
	}
}
