package biz.hammurapi.dataflow;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import biz.hammurapi.config.Component;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.util.Worker;

/**
 * Performs multi-threaded data pumping between data sources and 
 * sinks. Converts data item types as necessary.
 * @author Pavel
 */
public class DataPump implements Component {

	public void setOwner(Object owner) {
		// Nothing to do.		
	}
	
	private Worker worker;
	
	/**
	 * Worker to execute pumping jobs. If worker is not set, all tasks are 
	 * executed in the starting thread.
	 * @param worker
	 */
	public void setWorker(Worker worker) {
		this.worker = worker;
	}
	
	private class Connection {
		DataSource source;
		DataSinkProxy sink;
		
		public Connection(DataSource source, DataSinkProxy sink) {
			super();
			this.source = source;
			this.sink = sink;
		}				
	}
	
	private List connections = new ArrayList();
	
	/**
	 * Blocks sources when stopped.
	 * @author Pavel
	 *
	 */
	protected class DataSinkProxy implements DataSink {
		
		DataSink sink;
		private DataItemInfo[] sinkInfo;
		
		public DataSinkProxy(DataSink sink) {
			this.sink = sink;
			this.sinkInfo = sink.getDataInfo();
		}

		public boolean addData(final Data data) {
			if (sink.controlsMultiplexing()) {
				return sink.addData(passData(data, sinkInfo));
			} 
			
			Runnable pumpJob = new Runnable() {

				public void run() {
					sink.addData(passData(data, sinkInfo));
				}
				
			};
			
			postJob(pumpJob);			
			return false;
		}

		public boolean controlsMultiplexing() {
			return sink.controlsMultiplexing();
		}

		public DataItemInfo[] getDataInfo() {
			return sink.getDataInfo();
		}
		
	}
	
	private boolean isStarted = false;
	
	/**
	 * Adds connection to be processed.
	 * @param source
	 * @param sink
	 */
	public void addConnection(DataSource source, DataSink sink) {
		synchronized (connections) {
			connections.add(new Connection(source, new DataSinkProxy(sink)));
		}
	}
	
	/**
	 * Invoked in start() to establish connection between source and sink.
	 * @param source
	 * @param sink
	 */
	protected void connect(DataSource source, DataSinkProxy sink) {
		source.setSink(new DataSinkProxy(sink));
	}

	/**
	 * Invoked in stop() to terminate connection between source and sink.
	 * This implementation invokes source.setSink(null);
	 * @param source
	 * @param sink
	 */
	protected void disconnect(DataSource source, DataSinkProxy sink) {
		source.setSink(null);
	}
	
	/**
	 * Starts data pumping.
	 */
	public void start() {
		synchronized (connections) {
			if (!isStarted) {
				isStarted = true;
				Iterator cit = connections.iterator();
				while (cit.hasNext()) {
					Connection con = (Connection) cit.next();
					connect(con.source, con.sink);
				}
			}
		}		
	}

	/**
	 * Stops/pauses data pumping. Start method can be invoked again after
	 * stop to resume data pumping. 
	 */
	public void stop() {
		synchronized (connections) {
			if (isStarted) {
				isStarted = false;
				Iterator cit = connections.iterator();
				while (cit.hasNext()) {
					Connection con = (Connection) cit.next();
					disconnect(con.source, con.sink);
				}
			}
		}		
	}
	
	/**
	 * Passes data from source to sink adjusting data types. If not conversion
	 * is required data is returned as-is.
	 * Override for logging, etc.
	 * @param data
	 * @param sinkInfo 
	 * @return
	 */
	protected Data passData(Data data, DataItemInfo[] sinkInfo) {
		if (sinkInfo==null || sinkInfo.length==0) {
			return data;
		}
		
		Map dataItemsWithAdjustedTypes = new HashMap();
		for (int i=0; i<sinkInfo.length; ++i) {
			Object inputItem = data.get(sinkInfo[i].getName());
			if (inputItem!=null && !sinkInfo[i].getType().isInstance(inputItem)) {
				dataItemsWithAdjustedTypes.put(sinkInfo[i].getName(), convert(inputItem, sinkInfo[i].getType()));
			}
		}
		
		return dataItemsWithAdjustedTypes.isEmpty() ? data : createData(dataItemsWithAdjustedTypes , data);
	}

	/**
	 * Converts data item to expected target type.
	 * This implementation uses ConvertingService. 
	 * Override as needed.
	 * @param source
	 * @param targetType
	 * @return
	 */
	protected Object convert(Object source, Class targetType) {
		return ConvertingService.convert(source, targetType);
	}

	/**
	 * Creates new data with return values or error. This implementation creates
	 * Map data.
	 * @param newValues
	 * @param chain
	 * @return
	 */
	protected Data createData(Map newValues, Data chain) {
		return chain==null ? new MapData(newValues) : new MapData(newValues, new Data[] {chain});
	}
	
	/**
	 * This method block until all pumping jobs finish execution.
	 * If worker is not set, this method returns immediately.
	 * @throws InterruptedException 
	 */
	public void join() throws InterruptedException {
		if (worker!=null) {
			synchronized (jobCounter) {
				while (jobCounter[0]>0) {
					jobCounter.wait();
				}
			}
		}
	}
	
	// Counts active jobs.
	private int[] jobCounter = {0};

	/**
	 * Executes job either in the current thread or in a worker thread.
	 * @param job
	 */
	protected void postJob(final Runnable job) {
		
		Runnable workerJob = new Runnable() {

			public void run() {
				try {
					job.run();
				} finally {
					synchronized (jobCounter) {
						if (--jobCounter[0] <= 0) {
							jobCounter.notifyAll();
						}
					}
				}					
			}
			
		};
		
		synchronized (jobCounter) {
			++jobCounter[0];
		}
		
		if (worker==null) {
			workerJob.run();
		} else {			
			if (!worker.post(workerJob)) {
				workerJob.run();
			}
		}
	}
}
