package biz.hammurapi.dataflow;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.util.ClassLoaderRepository;

import biz.hammurapi.convert.ConvertingService;

/**
 * Data pipe wrapper for Java method.
 * Input data items are named after method parameters. Wrapped method shall be compiled with
 * debug on for correct resolution of parameter names. If parameter names cannot 
 * be resolved, they are automatically named arg0, ...
 * @author Pavel
 *
 */
public class JavaMethodDataPipe implements DataPipe {
	
	public static final String ERROR = "error";
	public static final String RESULT = "result";

	private Method method;
	private Object instance;
	private String returnName;
	private boolean beanToData;
	private boolean iterateOverReturn;
	private String[] parameterNames;
	private Class[] parameterTypes;
	private DataSink sink;
	private DataItemInfo[] dataInfo;
	
	/**
	 * Creates data pipe which invokes given method.
	 * @param method Method.
	 * @param instance Instances to invoke method against.
	 * @param returnName If beanToData is false, return value is stored in returnName data item. If beanToData is true,
	 * return name prefixes bean property names, unless it is null. If beanToData is false and returnName is null, then
	 * returnName defaults to "result". If method return type is Data then values from it are copied to output prefixed with
	 * return name or as-is if return name is null.
	 * @param beanToData If true return value Java Bean is converted to Data using BeanData class.
	 * @param iterateOverReturn If true and return value is collection or iterator, 
	 * then the return value is iterated over.
	 */
	public JavaMethodDataPipe(
			Method method, 
			Object instance, 
			String returnName,
			boolean beanToData, 			
			boolean iterateOverReturn) 
	{
		this.method = method;
		this.instance = instance;
		this.returnName = returnName;
		this.beanToData = beanToData;
		this.iterateOverReturn = iterateOverReturn;
		parameterTypes = method.getParameterTypes();
		parameterNames = new String[parameterTypes.length];
		for (int i=0; i<parameterTypes.length; ++i) {
			parameterNames[i] = "arg"+i;
		}
		Class declaringClass = method.getDeclaringClass();
		ClassLoaderRepository clr = new ClassLoaderRepository(declaringClass.getClassLoader());
		try {
			JavaClass dcl = clr.loadClass(declaringClass);
			org.apache.bcel.classfile.Method mth = dcl.getMethod(method);
			LocalVariableTable lvt = mth.getLocalVariableTable();
			LocalVariable[] localVars = lvt==null ? null : lvt.getLocalVariableTable();
			if (localVars != null) {
				for (int i=0, j = mth.isStatic() ? 0 : 1; i<parameterNames.length && j<localVars.length; ++i, ++j) {
					LocalVariable localVariable = localVars[j];
					if (localVariable!=null) {
						parameterNames[i] = localVariable.getName();
					}
				}
			}
		} catch (ClassNotFoundException e) {
			throw new DataFlowException("Should never happen: "+e, e);
		}
		dataInfo = new DataItemInfo[parameterTypes.length];
		for (int i=0; i<dataInfo.length; ++i) {
			final int idx = i;
			dataInfo[i] = new DataItemInfo() {

				public String getName() {
					return parameterNames[idx];
				}

				public Class getType() {
					return parameterTypes[idx];
				}

				public boolean isRequired() {
					return parameterTypes[idx].isPrimitive();
				}
				
			};
		}
	}
	
	public boolean addData(Data data) {
		Object[] args = new Object[parameterTypes.length];
		for (int i=0; i<args.length; ++i) {
			args[i] = convert(data.get(parameterNames[i]), parameterTypes[i]);
		}
		
		try {
			Object ret = method.invoke(instance, args);
			if (ret!=null && sink!=null) {
				if (iterateOverReturn) {
					if (ret instanceof Collection) {
						Iterator it = ((Collection) ret).iterator();
						while (it.hasNext()) {
							resultToSink(it.next(), data);
						}
					} else if (ret instanceof Iterator) {
						while (((Iterator) ret).hasNext()) {
							resultToSink(((Iterator) ret).next(), data);
						}
					} else {
						resultToSink(ret, data);
					}
				} else {
					resultToSink(ret, data);					
				}
			}
		} catch (IllegalArgumentException e) {
			handleException(e, data);
		} catch (InvocationTargetException e) {
			Throwable cause = e.getCause();
			handleException(cause, data);
		} catch (Exception e) {
			throw new DataFlowException(e);
		}
		return false;
	}	

	private void resultToSink(Object next, Data input) throws 
		IntrospectionException, 
		IllegalArgumentException, 
		IllegalAccessException, 
		InvocationTargetException {
		
		Map newValues = new HashMap();
		if (next instanceof Data) {
			Data retData = (Data) next;
			Iterator it = retData.getPropertyNames().iterator();
			while (it.hasNext()) {
				String pName = (String) it.next();
				newValues.put(returnName==null ? pName : returnName+pName, retData.get(pName));
			}
		} else if (beanToData) {
			BeanInfo beanInfo = Introspector.getBeanInfo(next.getClass());
			PropertyDescriptor[] pda = beanInfo.getPropertyDescriptors();
			for (int i=0; i<pda.length; ++i) {
				Method readMethod = pda[i].getReadMethod();
				if (readMethod!=null && readMethod.getParameterTypes().length==0) {
					newValues.put(returnName==null ? pda[i].getName() : returnName + pda[i].getName(), readMethod.invoke(next, null));
				}
			}
		} else {
			newValues.put(returnName==null ? RESULT : returnName, next);
		}
		sink.addData(createData(newValues, input));
	}

	private void handleException(Throwable cause, Data input) {
		Map.Entry handler = null;
		Iterator it = exceptionHandlers.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) it.next();
			Class exceptionType = (Class) entry.getKey();
			if (exceptionType.isInstance(cause)) {
				if (handler == null || ((Class) handler.getKey()).isAssignableFrom(exceptionType)) {
					handler = entry;
				}
			}
		}
		
		if (handler==null) {
			throw new DataFlowException("Unhandled exception", cause);
		}
		
		if (sink!=null) {
			Map newValues = new HashMap();
			newValues.put(ERROR, cause);
			sink.addData(createData(newValues, input));
		}
	}

	public boolean controlsMultiplexing() {
		return false;
	}

	public DataItemInfo[] getDataInfo() {
		return dataInfo;
	}

	public String getName() {
		return method.getName();
	}

	public void setSink(DataSink sink) {
		this.sink = sink;
	}
		
	private LinkedHashMap exceptionHandlers = new LinkedHashMap();

	/**
	 * Sets exception handler. Handlers are evaluated in the order of
	 * generality, i.e. handlers are never shadowed.
	 * @param exceptionType
	 * @param sink
	 */
	public void setExceptionHandler(Class exceptionType, DataSink sink) {
		exceptionHandlers.put(exceptionType, sink);
	}
	
	/**
	 * Converts input data item to target type. This implementation used
	 * ConvertingService, override as needed.
	 * @param src Source object
	 * @param targetType
	 * @return
	 */
	protected Object convert(Object src, Class targetType) {
		return ConvertingService.convert(src, targetType);
	}
	
	/**
	 * Creates new data with return values or error. This implementation creates
	 * Map data.
	 * @param newValues
	 * @param chain
	 * @return
	 */
	protected Data createData(Map newValues, Data chain) {
		return chain==null ? new MapData(newValues) : new MapData(newValues, new Data[] {chain});
	}
}
