package biz.hammurapi.dataflow;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Dispatches data from source to sinks in the order of addition.
 * This class doesn't do data type matching/conversion. Data info
 * returned by this class is aggregation of data info of connected sinks.
 * If several sinks have data items with the same name but different type, 
 * this class selects most specific data type. If it is not possible, it chooses
 * data type of data sink which was added to multiplexor earlier.  
 * @author Pavel
 */
public class Multiplexor implements DataSink {
	
	private List sinks = new ArrayList();
	
	private DataItemInfo[] dataInfo = {};

	private String name;

	/**
	 * Dispatched data to sinks. Always returns false (never consumes, even
	 * if one of sinks consumes).
	 */
	public boolean addData(Data data) {
		Iterator sit = sinks.iterator();
		while (sit.hasNext()) {
			if (((DataSink) data).addData(data)) {
				break;
			}
		}
		return false;
	}

	/**
	 * Always returns false.
	 */
	public boolean controlsMultiplexing() {
		return false;
	}

	/**
	 * Combines data input requirements from participating sinks.
	 */
	public DataItemInfo[] getDataInfo() {
		return dataInfo;
	}
	
	private DataItemInfo createDataItemInfo(final String name, final Class type, final boolean isRequired) {
		return new DataItemInfo() {

			public String getName() {
				return name;
			}

			public Class getType() {
				return type;
			}

			public boolean isRequired() {
				return isRequired;
			}
			
		};
	}

	public synchronized void addSink(DataSink sink) {
		sinks.add(sink);
		List newDataInfo = new ArrayList();
		
		DataItemInfo[] sinkDataInfo = sink.getDataInfo();
		Z: for (int sinkIdx=0; sinkIdx<sinkDataInfo.length; ++sinkIdx) {
			DataItemInfo sinkItem = sinkDataInfo[sinkIdx];
			for (int multiplexorIdx = 0; multiplexorIdx<dataInfo.length; ++multiplexorIdx) {
				DataItemInfo multiplexorItem = dataInfo[multiplexorIdx];
				if (sinkItem.getName().equals(multiplexorItem.getName())) {
					if (sinkItem.getType().isAssignableFrom(multiplexorItem.getType())) { // Multiplexor is more specific
						if (sinkItem.isRequired() && !multiplexorItem.isRequired()) {
							dataInfo[multiplexorIdx] = null;
							newDataInfo.add(createDataItemInfo(sinkItem.getName(), multiplexorItem.getType(), true));							
						}
					} else if (multiplexorItem.getType().isAssignableFrom(sinkItem.getType())) { // Sink is more specific
						dataInfo[multiplexorIdx] = null;
						newDataInfo.add(createDataItemInfo(sinkItem.getName(), sinkItem.getType(), sinkItem.isRequired() || multiplexorItem.isRequired()));													
					} else { // Unrelated, retain multiplexor's.
						if (sinkItem.isRequired() && !multiplexorItem.isRequired()) {
							dataInfo[multiplexorIdx] = null;
							newDataInfo.add(createDataItemInfo(sinkItem.getName(), multiplexorItem.getType(), true));							
						}						
					}
					continue Z;
				}
			}
			newDataInfo.add(sinkItem);
		}

		for (int i=0; i<dataInfo.length; ++i) {
			if (dataInfo[i]!=null) {
				newDataInfo.add(dataInfo[i]);
			}
		}

		dataInfo = (DataItemInfo[]) newDataInfo.toArray(new DataItemInfo[newDataInfo.size()]);
	}

	public String getName() {
		return name;
	}
	
	public void setName(String name) {
		this.name = name;
	}

}
