/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.dispatch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import biz.hammurapi.convert.DuckConverterFactory;

/**
 * This class dispatches objects to invocation handlers which can accept them.
 * Target invocation handlers are organized in "buckets" keyed by the argument class. This makes dispatching 
 * very efficient at runtime because only compatible handlers are interated over for invocation.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class Dispatcher {
	private List handlers=new ArrayList();
	private Map buckets=new HashMap();
	
	/**
	 * Creates dispatcher.
	 * @param targets Collection which contains instances of InvocationTarget or InvocationHandler. 
	 */
	public Dispatcher(Collection targets) {
		super();
		Set allClasses=new HashSet();
		Iterator it=targets.iterator();
		while (it.hasNext()) {
			Object o=it.next();
			if (o instanceof InvocationTarget) {
				Iterator hit = ((InvocationTarget) o).getInvocationHandlers().iterator();
				while (hit.hasNext()) {
					InvocationHandler ih=(InvocationHandler) hit.next();
					this.handlers.add(ih);
					if (ih.getParameterType()!=null) {
						allClasses.add(ih.getParameterType());
					}
				}
			} else if (o instanceof InvocationHandler) {
				InvocationHandler handler=(InvocationHandler) o;
				this.handlers.add(handler);
				if (handler.getParameterType()!=null) {
					allClasses.add(handler.getParameterType());
				}
			} else {
				throw new IllegalArgumentException("Invalid target: "+o);
			}				
		}

		// Create buckets for known classes
		it=allClasses.iterator();
		while (it.hasNext()) {
			getClassBucket((Class) it.next());
		}
	}
	
	private ResultConsumer resultConsumer = new ResultConsumer() {
		public void consume(Object result) {
			if (result!=null) {
				Dispatcher.this.consumeResult(result);
			}
		}
	};
	
	/**
	 * Simple form of invocation used by dispatcher
	 * to chain invocations.
	 * @author Pavel Vlasov
	 * @revision $Revision$
	 */
	protected interface Invocation {
		void invoke(Object arg, ResultConsumer resultConsumer);
	}

	/**
	 * Dispatches object to invocation handlers.
	 * Exceptions thrown by handlers are wrapped into DispatchException and that exception is dispatched as 
	 * though it is a return value. It is recommended to have an invocation handler which would handle thrown
	 * exceptions, e.g. log them. Errors are propagated up.
	 * @param arg Instance to be dispatched.
	 */
	public void dispatch(Object arg) {
		if (arg!=null) {
			getClassBucket(arg.getClass()).invoke(arg, resultConsumer);
		}		
	}
	
	/**
	 * @param clazz
	 * @return Invocation which contains handlers compatible with given class. Handlers are ordered by inverse affinity. I.e. handlers with less 
	 * specific parameters are invoked before handlers with more specific parameters. Handlers with unknown parameter type are invoked first.
	 */
	private Invocation getClassBucket(final Class clazz) {
		synchronized (buckets) {
			Invocation ret=(Invocation) buckets.get(clazz);
			if (ret==null) {
				class CompatibleEntry implements Comparable {
					Integer affinity;
					InvocationHandler handler;
					
					CompatibleEntry(Integer affinity, InvocationHandler handler) {
						super();
						this.affinity = affinity;
						this.handler = handler;
					}

					/**
					 * Higher hash code (affinity) comes first
					 */
					public int compareTo(Object obj) {
						if (obj==this) {
							return 0;
						}
						return obj.hashCode() - hashCode();
					}
					
					public int hashCode() {
						return affinity==null ? super.hashCode() : affinity.intValue();
					}										
				}
				
				List compatible=new ArrayList();
				Iterator it=handlers.iterator();
				while (it.hasNext()) {
					InvocationHandler handler=(InvocationHandler) it.next();
					if (handler.getParameterType()==null || handler.getParameterType().isAssignableFrom(clazz)) {
						Integer affinity = handler.getParameterType()==null ? new Integer(Integer.MAX_VALUE) : DuckConverterFactory.classAffinity(clazz, handler.getParameterType());
						compatible.add(new CompatibleEntry(affinity, handler));
					}
				}
				
				Collections.sort(compatible);
				
				final InvocationHandler[] bucketHandlers=new InvocationHandler[compatible.size()];
				Iterator cit = compatible.iterator();
				for (int i=0; cit.hasNext(); ++i) {
					bucketHandlers[i] = ((CompatibleEntry) cit.next()).handler;
				}
				
				ret = new Invocation() {

					public void invoke(Object arg, ResultConsumer resultConsumer) {
						for (int i=0, l=bucketHandlers.length; i<l; i++) {
							try {
								bucketHandlers[i].invoke(arg, resultConsumer);
							} catch (Exception e) {
								dispatch(new DispatchException(bucketHandlers[i], e));
							} catch (Error e) {
								throw e;
							} catch (Throwable e) {
								throw new RuntimeException("This should never happen", e);
							}
						}
					}
					
					public String toString() {
						return "[class bucket] "+clazz.getName()+", "+bucketHandlers.length+" handlers";
					}
					
				};
				
				buckets.put(clazz, ret);
			}
			return ret;			
		}
	}

	/**
	 * Consumes invocation results. This implementation
	 * simply dispatches results. 
	 * Override this method to provide more advanced results handling.
	 * @param result
	 */
	protected void consumeResult(Object result) {
		dispatch(result);
	}
	
	/**
	 * @return Number of handlers
	 */
	public int size() {
		return handlers.size();
	}
}
