/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.dispatch;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * This class introspects an object passed to the constructor and creates
 * invocation handlers for methods with one argument and the name provided in
 * the constructor. 
 * Invocation handler would not invoke target method with incompatible argument.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class IntrospectingInvocationTarget implements InvocationTarget {

	List handlers=new ArrayList();
	
	/**
	 * Introspects the target and creates an array of invocation handlers. 
	 * Invocation handlers created by introspection report their return types as
	 * fact types. getFactTypes() returns empty array if return type is void.
	 * @param targetInstance Target instance
	 * @param methodName Method name
	 * @param parameterType Invocation handlers shall accept arguments of this type. Can be null. 
	 */
	public IntrospectingInvocationTarget(final Object targetInstance, final String methodName, Class parameterType) {
		super();
		
		
		Method[] methods=targetInstance.getClass().getMethods();			
		for (int i=0; i<methods.length; i++) {
			if (methods[i].getParameterTypes().length == 1) {
				if (methodName.equals(methods[i].getName())) {
					final Method method=methods[i];
					final Class actualParameterType = method.getParameterTypes()[0];
					
					// Make sure that target method will accept parameterType arguments
					if ((parameterType==null || actualParameterType.isAssignableFrom(parameterType))) {
						handlers.add(
								new InvocationHandler() {
	
									public void invoke(Object arg, ResultConsumer resultConsumer) throws Throwable {
										// Invoke only with compatible parameters.
										if (actualParameterType.isInstance(arg)) {
											Object ret=method.invoke(targetInstance, new Object[] {arg});
											if (ret!=null && resultConsumer!=null) {
												resultConsumer.consume(ret);
											}
										}
									}
									
									public Class getParameterType() {
										return actualParameterType;
									}
									
									public String toString() {
										return "["+methodName+" handler] Target method: "+method;
									}

									public Class[] getFactTypes() {
										Class ret = method.getReturnType();
										return void.class.equals(ret) ? new Class[] {} : new Class[] {ret};
									}
									
								});
					}
				}
			}
		}		
		
	}

	public Collection getInvocationHandlers() {
		return Collections.unmodifiableCollection(handlers);
	}

}
