/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.dispatch;

import java.util.Collection;
import java.util.LinkedList;

import biz.hammurapi.util.Worker;

/**
 * This class puts results to a queue, which is processed by a worker.
 * An internal thread is used to push jobs to worker or execute jobs if worker is null.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class QueuingDispatcher extends Dispatcher {

	private Thread queueProcessor;
	
	/**
	 * Constructor
	 * @param targets
	 * @param worker Worker to delegate dispatching jobs to. Can be null.
	 */
	public QueuingDispatcher(Collection targets, final Worker worker) {
		super(targets);
		
		queueProcessor=new Thread() {
			public void run() {
				while (true) {
					Runnable job=null;
					synchronized (queue) {
						while (!shutdown && queue.isEmpty()) {
							try {
								queue.wait();
							} catch (InterruptedException e) {
								shutdown=true;
								queue.clear();
								e.printStackTrace();
								return;
							}
						}
						
						if (shutdown && queue.isEmpty()) {
							return;
						}
						
						job=(Runnable) queue.removeFirst();
					}
					
					// Try to post job to worker, execute the job if worker is null or doesn't accept the job.
					if (worker==null || !worker.post(job)) {
						job.run();
					}
				}
			}
		};
		
		queueProcessor.setName("[Dispatch Queue Processor] "+queueProcessor.getName());
		queueProcessor.start();
	}

	protected LinkedList queue=new LinkedList();
	private int[] jobCounter={0};
	
	/**
	 * Allows subclasses to post non-dispatching jobs to the queue.
	 * @param job
	 */
	protected void postJobToQueue(final Runnable foreignJob) {
		synchronized (queue) {			
			Runnable job=new Runnable() {
				
				// Increment job counter.
				{
					synchronized (jobCounter) {
						++jobCounter[0];
					}

				}
				
				/**
				 * dispatch result and decrement job counter.
				 */
				public void run() {
					try {
						foreignJob.run();
					} finally {
						synchronized (jobCounter) {
							--jobCounter[0];
							if (jobCounter[0]==0) {
								jobCounter.notifyAll();
							}
						}
					}
				}
				
				public String toString() {
					return "[foreign job wrapper] "+foreignJob;
				}
			};
			
			// Add job to queue and notify posting thread.
			queue.add(job);
			queue.notifyAll();
		}		
	}
	
	protected class DispatchJob implements Runnable {
		private Object payload;
		
		public Object getPayload() {
			return payload;
		}
		
		public DispatchJob(Object payload) {
			this.payload=payload;
			
			synchronized (jobCounter) {
				++jobCounter[0];
			}
		}
		
		public void run() {
			try {
				QueuingDispatcher.super.dispatch(payload);
			} finally {
				done();
			}
		}

		public void done() {
			synchronized (jobCounter) {
				--jobCounter[0];
				if (jobCounter[0]==0) {
					jobCounter.notifyAll();
				}
			}
		}
		
		public String toString() {
			return "[dispatch job] "+payload;
		}
		
	}
	
	/**
	 * Puts argument to dispatching queue.
	 * The queue is processed by an internal thread, which posts dispatching jobs to worker or
	 * processes them itself if worker is null.
	 */
	public void dispatch(final Object arg) {
		if (arg!=null) {
			if (shutdown) {
				throw new IllegalStateException("Dispatcher is shut down");
			}
			synchronized (queue) {			
				// Add job to queue and notify posting thread.
				queue.add(new DispatchJob(arg));
				queue.notifyAll();
			}
		}
	}
	
	/**
	 * Blocks until all jobs are processed.
	 * @throws InterruptedException 
	 */
	public void join() throws InterruptedException {
		synchronized (jobCounter) {
			while (jobCounter[0]!=0) {
				jobCounter.wait();
			}
		}
	}
	
	private boolean shutdown=false;
	
	/**
	 * Processes all jobs and stops queue processing thread.
	 * @throws InterruptedException 
	 *
	 */
	public void stop() throws InterruptedException {
		shutdown=true;
		//join();
		
		// Wakes up processing thread if queue is empty. 
		synchronized (queue) {
			if (queue.isEmpty()) {
				queue.notifyAll();
			}
		}
		queueProcessor.join();

	}
}
