/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.rules;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.tree.TreeNode;

import org.w3c.dom.Element;

import biz.hammurapi.config.Context;
import biz.hammurapi.config.PropertyParser;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.swing.LazyTreeNode;
import biz.hammurapi.swing.LazyTreeNodeTableVisualizable;
import biz.hammurapi.swing.Visualizable;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * Base class for conclusions.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class Conclusion 
implements 
	DomSerializable, 
	Serializable, 
	Visualizable, 
	Negatable,
	Supercedable {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -2364827338956686564L;
	
	/**
	 * Map is used to merge derivations without having to introduce
	 * side effects in equals().
	 */
	private Map derivations=new HashMap();
		
	/**
	 * Slots (other facts) belonging to this fact.
	 */
	protected HashMap slots=new HashMap();
	
	private int hashCode=getClass().getName().hashCode();

	private int depth;

	private String pattern;
	
	/**
	 * Puts value to a slot.
	 * Value gets converted to a handle.
	 * @param slotName
	 * @param value
	 */
	protected synchronized void setSlot(String slotName, Object value) {
		if (value!=null) {
			hashCode^=value.hashCode();
		}
		hashCode^=slotName.hashCode();
		Object oldValue = slots.put(slotName, value);
		if (oldValue!=null) {
			hashCode^=oldValue.hashCode();
		}
	}

	public int hashCode() {		
		return hashCode;
	}
	
	public boolean equals(Object obj) {		
		if (obj==this) {
			return true;
		}
		
		return obj!=null && 
				hashCode==obj.hashCode() && 
				getClass().equals(obj.getClass()) && 
				slots.equals(((Conclusion) obj).slots);
	}
	
	/**
	 * Merges derivations of two equal conclusions. This is useful when
	 * two there are two paths to the same conclusion.
	 * @param otherConclusion
	 */
	public void mergeDerivations(Conclusion otherConclusion) {
		if (this!=otherConclusion && equals(otherConclusion)) {
			// Merging derivations sets.
			Iterator it=otherConclusion.getDerivations().iterator();
			while (it.hasNext()) {
				Object o=it.next();
				derivations.put(o,o);
			}
			otherConclusion.derivations=new HashMap(derivations);			
		}
	}
		
	/**
	 * One conclusion supercedes another if it is a subclass of the other
	 * and slots of the both are equal. In other words more specific 
	 * conclusion supercedec more generic. E.g. conclusion that Mary is a mother of Joe 
	 * is more specific than that Mary is a parent of Joe.
	 * @param conclusion
	 * @return true if this fact is more specific than argument.
	 */
	public synchronized boolean supercedes(Object obj) {
		return obj instanceof Conclusion
			&& obj.getClass().isAssignableFrom(getClass()) 
			&& !getClass().isAssignableFrom(obj.getClass()) 
			&& slots.equals(((Conclusion) obj).slots); 
	}

	/**
	 * It is possible to come to the same conclusion through multiple
	 * inference paths.
	 * @return Set of derivations.
	 */
	public Set getDerivations() {
		return Collections.unmodifiableSet(derivations.keySet());
	}
	
	/**
	 * @return number of derivations.
	 */
	public int getCardinality() {
		return derivations.size();
	}
	
	/**
	 * @return minimum derivation depth, in other words shortest
	 * logical chain which led to this conclusion.
	 */
	public int getDepth() {
		return depth;
	}

	public synchronized void toDom(Element holder) {
		holder.setAttribute("type", getClass().getName());
		holder.setAttribute("depth", String.valueOf(depth));
		holder.setAttribute("hash-code", Integer.toString(hashCode, Character.MAX_RADIX));
				
		Iterator it=slots.keySet().iterator();
		while (it.hasNext()) {
			Element se=holder.getOwnerDocument().createElement("slot");
			holder.appendChild(se);			
			String slotName=(String) it.next();
			se.setAttribute("slot-name", slotName);
			DomSerializable sds = (DomSerializable) ConvertingService.convert(getSlot(slotName), DomSerializable.class);
			sds.toDom(se);
		}		
		
		it=derivations.keySet().iterator();
		while (it.hasNext()) {
			Derivation derivation=(Derivation) it.next();
			Element de=holder.getOwnerDocument().createElement("derivation");
			holder.appendChild(de);
			derivation.toDom(de);
		}
	}
	
	/**
	 * Returns true if negator negates this conclusion any of its slots or all its derivations.
	 * If conclusion has more than one derivation that negated derivations are removed from 
	 * derivations collection. Conclusion is negated based on derivations only if at leas one of its derivations
	 * is negated. In other words conclusion which collection of derivations is empty will not be negated based
	 * on derivations but only based on self and slots.
	 * @param negator
	 * @return
	 */
	public synchronized boolean isNegatedBy(Negator negator) {
		
		// Don't negate self.
		if (negator==this || this.equals(negator)) {
			return false;
		}
		
		// If negator negates this return true;
		if (negator.negates(this)) {
			return true;
		}
		
		// If negator negates any of slots return true;
		Iterator it=slots.values().iterator();
		while (it.hasNext()) {
			if (negator.negates(it.next())) {
				return true;
			}
		}
		
		// If conclusion is not derived then return false;
		if (derivations.isEmpty()) {
			return false;
		}
		
		// If all derivations are negated return true
		it=derivations.keySet().iterator();
		while (it.hasNext()) {
			Derivation d=(Derivation) it.next();
			if (!d.negatedBy(negator)) {
				return false; // If there is derivation which is not negated return false;
			}
		}
		
		return true; // All derivations were negated
	}
		
	/**
	 * Adds a derivation to conclusion.
	 * @param derivation
	 */
	public synchronized void addDerivation(Derivation derivation) {
		if (derivation!=null) {
			derivations.put(derivation, derivation);			
			depth = (derivations.size()==1 ? derivation.getDepth() : Math.min(depth, derivation.getDepth()))+1;
		}
	}

	/**
	 * Returns slot value
	 */
	protected Object getSlot(String name) {
		return slots.get(name);
	}	

	/**
	 * If pattern is null then outputs class name and list of slots and their values,
	 * otherwise formats pattern.
	 */
	public String toString() {
		if (pattern==null) {
			StringBuffer ret=new StringBuffer("["+getClass().getName()+"] ");
			Iterator it=slots.entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry entry=(Entry) it.next();
				ret.append(entry.getKey());
				ret.append("=");
				ret.append(entry.getValue());
				if (it.hasNext()) {
					ret.append(", ");
				}
			}
			return ret.toString();
		}
		
		return new PropertyParser(
				new Context() {
					public Object get(String name) {
						return getSlot(name);
					}
				}, false).parse(pattern);
	}
	
	/**
	 * @param pattern Pattern to use in toString(). The pattern should use ${<code>slot name</code>} placeholders.
	 * E.g. Parent conclusion can use pattern "<code>${parent} is parent of ${child}</code>".  
	 */
	protected Conclusion(String pattern) {
		this.pattern=pattern;
	}
	
	/**
	 * Default constructor
	 */
	public Conclusion() {
		// Default constructor.
	}
	
	/**
	 * Convenience method to properly negate objects taking implementations of
	 * Negatable into account.
	 * @param o Object to be negated
	 * @param n Negator
	 * @return true if negator negates object
	 */
	public static boolean object2Negator(Object o, Negator n) {
		return o instanceof Negatable ? ((Negatable) o).isNegatedBy(n) : n.negates(o);
	}
	
	/**
	 * 
	 * @param fact
	 * @return True if the fact equals to one of slots or one of derivations
	 * is based on this fact. 
	 */
	public boolean isDerivedFrom(Object fact) {
		if (fact==null) {
			return false;
		}
		
		// If fact equalsany of slots return true;
		Iterator it=slots.values().iterator();
		while (it.hasNext()) {
			if (fact.equals(it.next())) {
				return true;
			}
		}
		
		// If any of derivations is derived from the fact return true
		it=derivations.keySet().iterator();
		while (it.hasNext()) {
			Derivation d=(Derivation) it.next();
			if (d.isDerivedFrom(fact)) {
				return true;
			}
		}
		
		return false; 		
	}

	public TreeNode toTreeNode(TreeNode parent, String title) {
		return new LazyTreeNodeTableVisualizable(parent, title + " [" + Conclusion.this.getClass().getName() + "] "+ Conclusion.this){
		
			public TableModel toTable() {
				DefaultTableModel tm=new DefaultTableModel(slots.size()+1,3);
				tm.setColumnIdentifiers(new String[] {"Slot", "Type", "Value"});

				tm.setValueAt("(this)", 0, 0);
				tm.setValueAt(Conclusion.this.getClass().getName(), 0, 1);
				tm.setValueAt(Conclusion.this, 0, 2);
				
				Iterator it=slots.keySet().iterator();
				for (int i=1; it.hasNext(); i++) {
					String key=(String) it.next();
					Object slot=getSlot(key);
					tm.setValueAt(key, i, 0);
					tm.setValueAt(slot.getClass().getName(), i, 1);
					tm.setValueAt(slot, i, 2);
				}		
				
				return tm;
			}
		
			protected List loadChildren() {
				List ret = new ArrayList();
				
				ret.add(new LazyTreeNode(this, "Slots") {

					protected List loadChildren() {
						List ret = new ArrayList();
						Iterator it=slots.keySet().iterator();
						while (it.hasNext()) {
							String key=(String) it.next();
							Visualizable svs = (Visualizable) ConvertingService.convert(getSlot(key), Visualizable.class); 
							ret.add(svs.toTreeNode(this, key));
						}
						return ret;
					}
					
				});
				
				if (!derivations.isEmpty()) {
					ret.add(new LazyTreeNode(this, "Derivations") {

						protected List loadChildren() {
							List ret = new ArrayList();
							Iterator it=derivations.keySet().iterator();
							while (it.hasNext()) {
								Visualizable dvs = (Visualizable) ConvertingService.convert(it.next(), Visualizable.class); 
								ret.add(dvs.toTreeNode(this, ""));
							}
							return ret;
						}
						
					});
				}
				return ret;
			}
		}; 
	}
	
}
