/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.rules;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.tree.TreeNode;

import org.w3c.dom.Element;

import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.swing.LazyTreeNodeTableVisualizable;
import biz.hammurapi.swing.Visualizable;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * This class is a tracking element for inference process.
 * It is created by rules during inference process and attached to facts.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class Derivation implements DomSerializable, Serializable, Visualizable {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 2217441881574645228L;

	// Class and method name and object hash code
	private String ruleFullName;
	private Object ruleInstance;
	private Method ruleMethod;
	
	public Object getRuleInstance() {
		return ruleInstance;
	}
	
	public Method getRuleMethod() {
		return ruleMethod;
	}
		
	private int depth;
	
	/**
	 * Derivation depth is the size of the longest of logical chains which lead to this derivation.
	 * @return derivation depth
	 */
	public int getDepth() {
		return depth;
	}
	
	/**
	 * @param target
	 */
	public Derivation(Object target) {
		StringBuffer ruleBuf=new StringBuffer(target.getClass().getName());
		ruleBuf.append("[");
		if (target instanceof AbstractRule && ((AbstractRule) target).getName()!=null) {
			ruleBuf.append(((AbstractRule) target).getName());
		} else {
			ruleBuf.append(Integer.toString(target.hashCode(), Character.MAX_RADIX));
		}
		
		this.ruleInstance=target;
		ruleFullName=ruleBuf.toString();
		
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + depth;
		result = PRIME * result + ((facts == null) ? 0 : facts.hashCode());
		result = PRIME * result + hashCode;
		result = PRIME * result + ((ruleFullName == null) ? 0 : ruleFullName.hashCode());

		hashCode=result;
	}
	

	/**
	 * @param target
	 * @param method
	 */
	Derivation(Object target, Method method) {
		StringBuffer ruleBuf=new StringBuffer(target.getClass().getName());
		ruleBuf.append("[");
		if (target instanceof AbstractRule && ((AbstractRule) target).getName()!=null) {
			ruleBuf.append(((AbstractRule) target).getName());
		} else {
			ruleBuf.append(Integer.toString(target.hashCode(), Character.MAX_RADIX));
		}
		
		this.ruleInstance=target;
		this.ruleMethod=method;
		
		ruleBuf.append("].");
		ruleBuf.append(method.getName());
		ruleBuf.append("(");
		Class[] parameterTypes = method.getParameterTypes();
		for (int i=0; i<parameterTypes.length; i++) {
			if (i>0) {
				ruleBuf.append(",");
			}
			ruleBuf.append(parameterTypes[i].getName());
		}
		ruleBuf.append(")");
		ruleFullName=ruleBuf.toString();
		
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + depth;
		result = PRIME * result + ((facts == null) ? 0 : facts.hashCode());
		result = PRIME * result + hashCode;
		result = PRIME * result + ((ruleFullName == null) ? 0 : ruleFullName.hashCode());

		hashCode=result;
	}
	
	private Set facts=new HashSet();

	/**
	 * Adds fact to derivation.
	 * If fact is instance of conclusion then its derivation depth becomes Max(ownDepth, conclusionDepth).
	 * @param fact
	 */
	public synchronized void addSourceFact(Object fact) {		
		if (facts.add(fact)) {
			hashCode^=fact.hashCode();
			if (fact instanceof Conclusion) {
				depth=Math.max(depth, ((Conclusion) fact).getDepth());
			}
		}
	}
	
	/**
	 * @return Full rule name which includes class and method names and object hash code
	 */
	public String getRuleFullName() {
		return ruleFullName;
	}
	
	public synchronized String toString() {
		StringBuffer buf=new StringBuffer("[Derivation] "+ruleFullName);
		buf.append("[");
		Iterator it=facts.iterator();
		while (it.hasNext()) {
			buf.append(it.next());
			if (it.hasNext()) {
				buf.append(",");
			}
		}
		buf.append("]");
		return buf.toString();
	}

	public void toDom(Element holder) {
		holder.setAttribute("depth", String.valueOf(getDepth()));
		holder.setAttribute("rule", ruleFullName);
		holder.setAttribute("hash-code", Integer.toString(hashCode, Character.MAX_RADIX));
		Iterator it=facts.iterator();
		while (it.hasNext()) {
			Element fe=holder.getOwnerDocument().createElement("fact");
			holder.appendChild(fe);
			DomSerializable fds = (DomSerializable) ConvertingService.convert(it.next(), DomSerializable.class);
			fds.toDom(fe);
		}
	}
	
	/**
	 * Derivation is negated if any of its constituent facts in negated.
	 * @param negator
	 * @return
	 */
	public boolean negatedBy(Negator negator) {
		Iterator it=facts.iterator();
		while (it.hasNext()) {
			Object fact=it.next();
			if (fact instanceof Conclusion) {
				if (((Conclusion) fact).isNegatedBy(negator)) {
					return true;
				}
			} else {
				if (negator.negates(fact)) {
					return true;
				}
			}
		}
		return false;
	}
	
//	public boolean equals(Object obj) {
//		return obj==this || (
//				obj!=null 
//				&& hashCode==obj.hashCode() 
//				&& getClass().equals(obj.getClass()) 
//				&& rule.equals(((Derivation) obj).getRule()) 
//				&& facts.equals(((Derivation) obj).getFacts()));
//	}
	
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		
		if (obj == null) {
			return false;
		}
		
		if (getClass() != obj.getClass()) {
			return false;
		}
		
		final Derivation other = (Derivation) obj;
//		if (depth != other.depth) {
//			return false;
//		}
		
		if (facts == null) {
			if (other.facts != null) {
				return false;
			}
		} else if (!facts.equals(other.facts)) {
			return false;
		}
		
		if (ruleFullName == null) {
			if (other.ruleFullName != null) {
				return false;
			}			
		} else if (!ruleFullName.equals(other.ruleFullName)) {
			return false;
		}
		return true;
	}

	
	public Set getFacts() {
		return Collections.unmodifiableSet(facts);
	}

	private int hashCode;
	
	public int hashCode() {
		return hashCode;
	}

	public TreeNode toTreeNode(TreeNode parent, String title) {
		return new LazyTreeNodeTableVisualizable(parent, title + " [Derivation] " + ruleFullName +" [depth="+depth+"]") {

			protected List loadChildren() {
				List ret = new ArrayList();
				Iterator it=facts.iterator();
				while (it.hasNext()) {
					Visualizable fvs = (Visualizable) ConvertingService.convert(it.next(), Visualizable.class); 
					ret.add(fvs.toTreeNode(this, ""));
				}
				return ret;
			}

			public TableModel toTable() {
				DefaultTableModel tm=new DefaultTableModel(facts.size()+1,2);
				tm.setColumnIdentifiers(new String[] {"Type", "Value"});
				
				tm.setValueAt("(rule)", 0, 0);
				tm.setValueAt(ruleFullName, 0, 1);		
				
				Iterator it=facts.iterator();
				for (int i=1; it.hasNext(); i++) {
					Object fact=it.next();
					tm.setValueAt(fact.getClass().getName(), i, 0);
					tm.setValueAt(fact, i, 1);
				}		
				
				return tm;
			}
			
		};
	}

	public boolean isDerivedFrom(Object fact) {
		if (fact==null) {
			return false;
		}
		
		Iterator it=facts.iterator();
		while (it.hasNext()) {
			Object baseFact=it.next();
			if (fact.equals(baseFact)) {
				return true;
			}
			
			if (baseFact instanceof Conclusion && ((Conclusion) baseFact).isDerivedFrom(fact)) {
				return true;
			}
		}
		return false;
	}

}
