/**
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.rules;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.rules.Handle;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;

/**
 * Handle manager implementation which keeps handles in memory. 
 * Objects are compared by eqals().
 * This handle manager doesn't hold facts which superced each other, but retains only superceeding fact 
 * (e.g. if Mother supercedes Parent then Parent will not be retained in the manager when Mother is added). 
 * @author Pavel Vlasov
 * @version ${Revision}
 */
public class KnowledgeMaximizingHandleManager extends ComponentBase implements HandleManager, Component {
	
	private Map handleMap;

	private String storageReference;
	
	/**
	 * Path to object storage.
	 * @param storageReference
	 */
	public void setStorageReference(String storageReference) {
		this.storageReference = storageReference;
	}
	

	private static class HandleImpl implements Handle, Serializable {
		private static final long serialVersionUID = -6198273232968377370L;
		private Object master;
		
				
		HandleImpl(Object master) {
			this.master=master;
		}
		
		public boolean equals(Object obj) {
			return obj instanceof HandleImpl && master.equals(((HandleImpl) obj).master);
		}
		
		public int hashCode() {
			return master==null ? 0 : master.hashCode();
		}
		
		public String toString() {
			return getClass().getName() + " -> " + master;
		}
	}
	
	/**
	 * Adds new object to the manager if this object is not already there and there is no superceding object.
	 * If object being added supercedes any objects in the manager these objects will be removed.
	 */
	synchronized public Handle addObject(Object object) {
		Iterator it=handleMap.keySet().iterator();
		while (it.hasNext()) {
			HandleImpl existingHandle=(HandleImpl) it.next();
			if (existingHandle.master.equals(object)) {
				// Merge derivations for equal conclusions. 
				if (object!=existingHandle.master && object instanceof Conclusion) {
					((Conclusion) existingHandle.master).mergeDerivations((Conclusion) object);
				}
				return existingHandle;
			}
			
			if (object instanceof Supercedable && existingHandle.master instanceof Supercedable) {
				if (((Supercedable) existingHandle.master).supercedes(object)) {
					return existingHandle;
				}
			}
			
			if (object instanceof Supercedable && existingHandle.master instanceof Supercedable) {
				if (((Supercedable) object).supercedes(existingHandle.master)) {
					rebind(existingHandle, object);
					return existingHandle;
				}
			}			
		}
		
		Handle newHandle=new HandleImpl(object);
		handleMap.put(newHandle, newHandle);
		return newHandle;
	}

	synchronized public Object getObject(Handle handle) {
		if (handle instanceof HandleImpl) {
			return ((HandleImpl) handle).master;
		} 
		
		throw new IllegalArgumentException("Foreign handle: "+handle);
	}

	public Collection getObjects() {
		ArrayList ret=new ArrayList();
		Iterator it=handleMap.keySet().iterator();
		while (it.hasNext()) {
			ret.add(((HandleImpl) it.next()).master);
		}
		return ret;
	}

	public Collection getHandles() {
		return handleMap.keySet();
	}
	
	synchronized public void remove(Handle handle) {
		handleMap.remove(handle);
	}

	public boolean contains(Handle handle) {		
		return handleMap.containsKey(handle);
	}

	synchronized public void rebind(Handle handle, Object object) {
		if (handle instanceof HandleImpl) {
			handleMap.remove(handle);
			((HandleImpl) handle).master=object;
			handleMap.put(handle, handle);
		} else {
			throw new IllegalArgumentException("Foreign handle: "+handle);
		}
	}

	public void remove(Object obj) {
		handleMap.remove(new HandleImpl(obj));
	}

	public synchronized void clear() {
		handleMap.clear();
	}
	
	public void start() throws ConfigurationException {
		if (storageReference==null) {
			handleMap=new LinkedHashMap();
		} else {
			handleMap = (Map) ((ObjectStorage) get(storageReference)).get("handle-manager");
			if (handleMap==null) {
				handleMap=new HashMap();
				((ObjectStorage) get(storageReference)).put("handle-manager", handleMap);
			}
		}
	}

	public void stop() throws ConfigurationException {
		// Nothing
	}

	public synchronized boolean isNegatedBy(Negator negator) {
		Iterator it=handleMap.keySet().iterator();
		while (it.hasNext()) {
			HandleImpl handle=(HandleImpl) it.next();
			if (Conclusion.object2Negator(handle.master, negator)) {
				it.remove();
			}
		}
		
		return false;
	}
}
