/**
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.rules;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.RuntimeConfigurationException;

/**
 * Collection manager which holds collections in memory and 
 * can use storage to persist state. 
 * @author Pavel Vlasov
 * @version ${Revision}
 */
public class PojoCollectionManager extends ComponentBase implements Component, CollectionManager {

	private Map collectionMap;
	private Class collectionType = HashSet.class;
	private String storageReference;
	
	/**
	 * Path to object storage.
	 * @param storageReference
	 */
	public void setStorageReference(String storageReference) {
		this.storageReference = storageReference;
	}
	
	/**
	 * Starts collection manager. Loads collections from storage if storage reference is not null.
	 */
	synchronized public void start() throws ConfigurationException {		
		if (storageReference==null) {
			collectionMap=new HashMap();
		} else {
			collectionMap = (Map) ((ObjectStorage) get(storageReference)).get("collection-manager");
			if (collectionMap==null) {
				collectionMap=new HashMap();
				((ObjectStorage) get(storageReference)).put("collection-manager", collectionMap);
			}
		}
	}

	/**
	 * Empty method. Storage of collections is performed by object storage if storage reference is set.
	 */
	synchronized public void stop() throws ConfigurationException {
		// Nothing.
	}
		
	private class CollectionEntry {
		String ruleName;
		String collectionName;
		Object monitor;

		CollectionEntry(String ruleName, String collectionName, Object monitor) {
			this.ruleName = ruleName;
			this.collectionName = collectionName;
			this.monitor = monitor;
		}
		
		public String toString() {
			return "[Collection entry] Rule="+ruleName+", Collection="+collectionName;
		}
		
		public boolean equals(Object obj) {
			if (obj==this) {
				return true;
			}
			
			if (obj instanceof CollectionEntry) {
				CollectionEntry ce = (CollectionEntry) obj;
				
				return ruleName.equals(ce.ruleName) && collectionName.equals(ce.collectionName);
			}
			
			return false;
		}
		
		public int hashCode() {
			return ruleName.hashCode() ^ collectionName.hashCode();
		}
	}
	
	/**
	 * @see biz.hammurapi.rules.CollectionManager#getList(java.lang.String, java.lang.String)
	 */
	synchronized public Collection get(String ruleName, String listName, Object monitor) {
		CollectionEntry ce=new CollectionEntry(ruleName, listName, monitor);
		Collection ret=(Collection) collectionMap.get(ce);
		if (ret==null) {
			try {
				ret=(Collection) collectionType.newInstance();
				collectionMap.put(ce, ret);
			} catch (InstantiationException e) {
				throw new RuntimeConfigurationException("Cannot instantiate "+collectionType+": "+e, e);
			} catch (IllegalAccessException e) {
				throw new RuntimeConfigurationException("Cannot instantiate "+collectionType+": "+e, e);
			}  
		}
		return ret;
	}

	/* (non-Javadoc)
	 * @see biz.hammurapi.rules.CollectionManager#clear()
	 */
	public synchronized void clear() {
		Iterator it=collectionMap.values().iterator();
		while (it.hasNext()) {
			((Collection) it.next()).clear();
		}		
	}

	/**
	 * Iterates over all collections and removes negated facts.
	 * Collection operations are synchronized using monitor object passed in get() method for a given collection
	 * or collection itself if monitor is null.
	 */
	public synchronized boolean isNegatedBy(Negator negator) {
		Iterator it=collectionMap.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) it.next();
			CollectionEntry ce = (CollectionEntry) entry.getKey();
			Collection col = (Collection) entry.getValue();
			synchronized (ce.monitor==null ? col : ce.monitor) {
				Iterator cit=col.iterator();
				while (cit.hasNext()) {
					Object o=cit.next();
					if (o instanceof Conclusion) {
						if (((Conclusion) o).isNegatedBy(negator)) {
							cit.remove();
						}
					} else {
						if (negator.negates(o)) {
							cit.remove();
						}
					}
				}
			}
		}
		return false; // Return value has no meaning for collection manager.
	}
	
	public void setCollectionType(String collectionType) {
		try {
			this.collectionType = Class.forName(collectionType);
		} catch (ClassNotFoundException e) {
			throw new RuntimeConfigurationException("Cannot load collection class: "+e, e);
		}
	}

	public void setCollectionType(Class collectionType) {
		this.collectionType = collectionType;
	}
}
