/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.rules;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.dispatch.DispatchException;
import biz.hammurapi.dispatch.QueuingDispatcher;
import biz.hammurapi.util.Worker;


/**
 * This rules container uses QueuingDispatcher and negation semantics for remove().
 * It can also leverage worker (thread pool) provided in <code>worker-ref</code> attribute
 * for multithreaded inference.
 * This container also uses loop detection. 
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class QueueingRulesContainer extends RulesContainerBase implements Constants {
	
	private FactDispatcher dispatcher;
	private FactDispatcher removeDispatcher;
	private CollectionManager collectionManager;
	private HandleManager handleManager;
	private boolean doTracing;

	public QueueingRulesContainer() {
		super();
		doTracing = this instanceof ActionTracer;
	}

	/**
	 * Adds object to handle manager and dispatches to rules.
	 * If the object is instance of Negator then it is presented to collection manager.
	 */
	public void add(Object obj) {
		// Remove negators negated by this and then don't dispatch this if it is 
		// negated.
		if (negators!=null) {
			synchronized (negators) {
				// Remove negators negated by obj.
				if (obj instanceof Negator) {					
					Iterator it=negators.iterator();
					while (it.hasNext()) {
						if (Conclusion.object2Negator(it.next(), (Negator) obj)) {
							it.remove();
						}
					}
				}
				
				// Do not dispatch object if it is negated
				Iterator it=negators.iterator();
				while (it.hasNext()) {
					if (Conclusion.object2Negator(obj, (Negator) it.next())) {
						return;
					}
				}
				
			}
		}
		dispatcher.dispatch(obj);
	}
	
	/**
	 * Consumes DispatchException.
	 * This implementation just prints stack trace.
	 * Override if needed.
	 * @param exception
	 */
	protected void onDispatchException(DispatchException exception) {
		exception.printStackTrace();		
	}

	private Map recentFacts=new WeakHashMap();
	
	/**
	 * Checks recent conclusions and derivation depth to discard conclusions which came through dispatching too many times
	 * or conclusions with too big depth.  
	 * @param object
	 * @return true if there is no loop
	 */
	protected boolean checkLoop(Object object) {
		if (object instanceof Conclusion) {
			Conclusion cb=(Conclusion) object;
			
			synchronized (recentFacts) {
				Reference prevRef=(Reference) recentFacts.get(object);
				Conclusion prev = prevRef==null ? null : (Conclusion) prevRef.get();
				
				if (prev==null) {
					recentFacts.put(object, new WeakReference(object));
				} else {
					/**
					 * If conclusion recently passed dispatching then it is discarded, but
					 * derivations are merged if new conclusion is not derived from existing equal conclusion.
					 * This statement prevents logical loops like Parent -> Child -> Parent
					 */
					if (prev!=cb && !cb.isDerivedFrom(prev)) {
						cb.mergeDerivations(prev);
					}
					
					return false;
				}
			}
						
			/**
			 * If depth of the conclusion in more than 3 times number of handlers then such conclusion is
			 * discarded because most probably there is a logical loop. 
			 */
			if (cb.getDepth()>dispatcher.size()*3) {
				onDiscardedConclusion((Conclusion) object);
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Conclusions discarded because their derivation depth is too big are passed to this method.
	 * The method does nothing. Override as needed.
	 * @param conclusion
	 */
	protected void onDiscardedConclusion(Conclusion conclusion) {
		// Nothing - override if needed.
	}
	
	/**
	 * Removes object from handle manager and recent facts. Then creates a negator, negates facts in collection manager and 
	 * then dispatches the object to remove methods.
	 * Subclasses implementing ActionTracing undo actions instead of using a negator.
	 */
	public void remove(Object obj) {
		if (doTracing) {
			Collection objectActions = getObjectActions(obj);
			if (objectActions!=null) {
				Iterator it=objectActions.iterator();
				while (it.hasNext()) {
					((Action) it.next()).undo();
				}
			}
		} else {
			processNegator(newNegator(obj));
		}
		removeDispatcher.dispatch(obj);				
	}
	
	/**
	 * Returns actions performed when object was added to the database.
	 * This method must be overriden by subclasses which implement ActionTracer.
	 * @param obj
	 * @return
	 */
	protected Collection getObjectActions(Object obj) {
		throw new UnsupportedOperationException("Subclasses which implement ActionTracer must override this method");
	}

	/**
	 * Instantiates new negator.
	 * @param obj
	 * @return
	 */
	protected Negator newNegator(Object obj) {
		try {
			return (Negator) (negatorConstructor == null ? new EqualityNegator(obj) : negatorConstructor.newInstance(new Object[] {obj}));
		} catch (InstantiationException e) {
			throw new RuntimeException("Could not instantiate negator: "+e, e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Could not instantiate negator: "+e, e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException("Could not instantiate negator: "+e, e);
		}
	}
	
	private void processNegator(Negator negator) {
		// Remove from negators
		if (negators!=null) {
			synchronized (negators) {
				Iterator it=negators.iterator();
				while (it.hasNext()) {
					if (Conclusion.object2Negator(it.next(), negator)) {
						it.remove();
					}
				}
			}
		}
		
		synchronized (recentFacts) {
			Iterator it=recentFacts.keySet().iterator();
			while (it.hasNext()) {
				if (Conclusion.object2Negator(it.next(), negator)) {
					it.remove();
				}
			}
		}
		handleManager.isNegatedBy(negator);
		collectionManager.isNegatedBy(negator);
		dispatcher.isNegatedBy(negator);
		removeDispatcher.isNegatedBy(negator);
	}

	/**
	 * Invokes dispatcher's join() to wait until all 
	 * jobs are finished.
	 */
	public void executeRules() {
		try {
			dispatcher.join();
			removeDispatcher.join();
		} catch (InterruptedException e) {
			throw new RulesRuntimeException("Wait interrupted: "+e, e);
		}
	}
	
	private String workerRef;
	
	/**
	 * If this it true then negators posted to the bus are put to a collection and all new facts are checked
	 * against negators before being posted.
	 */
	private Collection negators;
	
	private Constructor negatorConstructor;
	
	public void configure(Node configNode, Context context, ClassLoader classLoader) throws ConfigurationException {
		super.configure(configNode, context, classLoader);
		Element ce=(Element) configNode;
		if (ce.hasAttribute(WORKER_REF)) {
			workerRef=ce.getAttribute(WORKER_REF);
		}
		
		if (ce.hasAttribute(RETAIN_NEGATORS) && "yes".equals(ce.getAttribute(RETAIN_NEGATORS))) {
			negators=new ArrayList();
		}
		
		if (ce.hasAttribute(NEGATOR_CLASS)) {
			try {
				Class negatorClass = Class.forName(ce.getAttribute(NEGATOR_CLASS));
				negatorConstructor = negatorClass.getConstructor(new Class[] {Object.class});
			} catch (ClassNotFoundException e) {
				throw new ConfigurationException("Negator class not found: "+e, e);
			} catch (SecurityException e) {
				throw new ConfigurationException("Cannot access negator constructor: "+e, e);
			} catch (NoSuchMethodException e) {
				throw new ConfigurationException("Negator constructor not found: "+e, e);
			}
		}
	}
	
	private class FactDispatcher extends QueuingDispatcher implements Negatable {
				
		/**
		 * @param targets
		 * @param worker
		 */
		public FactDispatcher(Collection targets, Worker worker) {
			super(targets, worker);
		}
		
		/**
		 * Dispatches object or executes knowledge base command.
		 */
		public void dispatch(Object arg) {
			if (arg instanceof KnowledgeBaseCommand) {
				((KnowledgeBaseCommand) arg).execute(QueueingRulesContainer.this);
			} else {
				super.dispatch(arg);
			}
		}

		/**
		 * Removes negated object from queue
		 */
		public boolean isNegatedBy(Negator negator) {
			synchronized (queue) {
				Iterator it=queue.iterator();
				while (it.hasNext()) {
					Object job = it.next();					
					if (job instanceof DispatchJob && Conclusion.object2Negator(((DispatchJob) job).getPayload(), negator)) {
						it.remove();
						((DispatchJob) job).done();
					}
				}
			}
			return false;
		}
			
	}
	
	public void start() throws ConfigurationException {
		super.start();
		collectionManager=(CollectionManager) get("/"+COLLECTION_MANAGER);
		handleManager=(HandleManager) get("/"+HANDLE_MANAGER);
		Worker worker = (Worker) (workerRef==null ? null : get(workerRef));
		
		dispatcher=new FactDispatcher(getComponents(), worker) {
			
			public void dispatch(final Object obj) {
				if (obj instanceof DispatchException) {
					onDispatchException((DispatchException) obj);
				} else if (checkLoop(obj)) {
					// Negate collections if negator
					if (obj instanceof Negator) {
						// Post negation job to queue. 
						postJobToQueue(
								new Runnable() {

									public void run() {
										processNegator((Negator) obj);
									}
									
								});
					}
					
					// Add only public facts to the handle manager.
					if (!(obj instanceof Fact && ((Fact) obj).isPrivate())) {
						handleManager.addObject(obj);
					}
					super.dispatch(obj);
				}
			}						
		};		
		Collection removeHandlers=new ArrayList();
		Iterator it=getComponents().iterator();
		while (it.hasNext()) {
			AbstractRule rule=(AbstractRule) it.next();
			removeHandlers.addAll(rule.getRemoveHandlers());
		}
		
		removeDispatcher=new FactDispatcher(removeHandlers, worker);
	}
	
	public void stop() throws ConfigurationException {
		try {
			dispatcher.stop();
		} catch (InterruptedException e) {
			throw new ConfigurationException("Could not stop dispatcher: "+e,e);
		}
		try {
			removeDispatcher.stop();
		} catch (InterruptedException e) {
			throw new ConfigurationException("Could not stop remove dispatcher: "+e,e);
		}
		super.stop();
	}

	public void reset() {
		Iterator it=getComponents().iterator();
		while (it.hasNext()) {
			((AbstractRule) it.next()).reset();
		}
	}

	public Collection getRules() {
		return Collections.unmodifiableCollection(getComponents());
	}		
}
