/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.rules;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.dispatch.InvocationHandler;
import biz.hammurapi.dispatch.ResultConsumer;
import biz.hammurapi.util.Observable;
import biz.hammurapi.util.Observer;
import biz.hammurapi.util.Versioned;


/**
 * Base class for rules.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class Rule extends AbstractRule {
	
	/**
	 * Interface to detect changes in arguments passed to inference methods.
	 * Default implementation detects changes in versioned and observable objects.
	 * @author Pavel Vlasov
	 * @revision $Revision$
	 */
	public interface ChangeDetector {
		boolean hasChanged();
	}
	
	/**
	 * Creates default change detector, which detects changes in versioned and observable objects.
	 * Subclasses can override this method to detect changes as appropriate for the application 
	 * domain.
	 * @param obj
	 * @return Change detector or null if change detection is not needed.
	 */
	protected ChangeDetector newChangeDetector(final Object obj) {
		return new ChangeDetector() {
			private int version;
			private boolean changed;
			
			private Observer observer=new Observer() {

				public void update(Observable observable, Object arg) {
					changed=true;
				}
				
			};
			
			{
				if (obj instanceof Versioned) {
					version=((Versioned) obj).getObjectVersion();
				} else if (obj instanceof Observable) {
					((Observable) obj).addObserver(observer);
				}
			}

			public boolean hasChanged() {
				if (obj instanceof Versioned) {
					return version!=((Versioned) obj).getObjectVersion();
				} else if (obj instanceof Observable) {
					((Observable) obj).removeObserver(observer);
					return changed;
				}
				return false;
			}
			
		};
	}
	
	/**
	 * Holds reference to the method currently being invoked.
	 * Used for injecting derivations in add() method.
	 */
	private ThreadLocal currentResultConsumerWrapper=new ThreadLocal();

	/**
	 * 
	 */
	private static final long serialVersionUID = 5320143875248203766L;
	
	private Collection removeHandlers=new ArrayList();
	private Collection addHandlers=new ArrayList();

	private String inferMethodName;

	private String removeMethodName;

	private String acceptMethodName;

	/**
	 * Each addtion to this class increments version number.
	 * @author Pavel Vlasov
	 * @revision $Revision$
	 */
	private class VersionedHashSet extends HashSet {
		private int version;
		
		public int getVersion() {
			return version;
		}
		
		public boolean add(Object arg) {
			if (super.add(arg)) {
				++version;
				return true;
			}
			
			return false;
		}
	}
	
	/**
	 * Default constructor.
	 * Uses "infer" for infer methods, "remove" for remove methods,
	 * and "accept" for filters.
	 *
	 */
	public Rule() {
		this("infer", "remove", "accept");
	}
		
	/**
	 * Instances of this interface are passed to <code>accept()</code> methods as second parameter. This parameter
	 * can be used if one <code>accept()</code> method is bound to several <code>join()</code> methods or to
	 * several <code>join()</code> method's parameters. 
	 * @author Pavel Vlasov
	 * @revision $Revision$
	 */
	public interface AcceptInfo {
		
		/**
		 * @return Parameter index in the target method.
		 */
		int parameterIndex();
		
		/**
		 * @return Target <code>join()</code> method.
		 */
		Method targetMethod();		
	}
	
	/**
	 * Creates handlers for join method
	 * @param method Join method.
	 * @param acceptMethods accept methods associated with this join method.
	 */
	void createJoinHandlers(final Method method, final Method[] acceptMethods) {
		Class[] parameterTypes = method.getParameterTypes();
		final Collection[] collections=new Collection[parameterTypes.length];
		
		StringBuffer msb=new StringBuffer(method.getName());
		msb.append("(");
		for (int i=0; i<parameterTypes.length; i++) {
			if (i>0) {
				msb.append(",");
			}
			msb.append(parameterTypes[i].getName());
		}
		msb.append(")");
		
		final String signature=msb.toString();

		for (int i=0; i<parameterTypes.length; i++) {
			collections[i]=getCollection(signature+"["+i+"]", collections);
		}
		
		for (int i=0; i<parameterTypes.length; i++) {				
			final int parameterIndex=i;
			final Class parameterType=parameterTypes[i];
			
			addHandlers.add(new InvocationHandler() {

				public void invoke(Object arg, ResultConsumer resultConsumer) throws Throwable {
					if (acceptMethods[parameterIndex]!=null) {
						AcceptInfo ai=new AcceptInfo() {

							public int parameterIndex() {
								return parameterIndex;
							}

							public Method targetMethod() {
								return method;
							}
							
							public String toString() {
								return "[AcceptInfo] "+method+", parameter index: "+parameterIndex;
							}
							
						};
						
						// Don't do anything if not accepted.
						if (Boolean.FALSE.equals(acceptMethods[parameterIndex].invoke(Rule.this, new Object[] {arg, ai}))) {
							return;
						}
					}
										
					Collection actions = doTracing ? new ArrayList() : null;
					
					synchronized (collections) {						
						// Add argument to its collection, run join if it was actually added
						if (collections[parameterIndex].add(arg)) {
							Object[] args=new Object[collections.length];							
							args[parameterIndex]=arg;
							VersionedHashSet negators = new VersionedHashSet(); // Negators posted in this iteration over collections.
							int[] negatorVersions = new int[collections.length];
							for (int i=0; i<collections.length; i++) {
								negatorVersions[i]=negators.getVersion();
							}
							
							doJoin(args, 0, negators, negatorVersions, actions, resultConsumer);
						}
					}
					
					if (doTracing && !actions.isEmpty()) {
						((ActionTracer) owner).addActions(actions);
					}
				}
				
				public String toString() {
					return "["+method.getName()+" handler] Target method: "+method+", target parameter: "+parameterIndex+" ("+parameterType+")";
				}

				/**
				 * @return last non-null argument.
				 */
				private int doJoin(Object[] args, int idx, VersionedHashSet negators, int[] negatorVersions, Collection actions, ResultConsumer rawConsumer) throws Throwable {
					if (idx==args.length) {												
						Object prevResultConsumerWrapper=currentResultConsumerWrapper.get();
						try {
							ResultConsumerWrapper resultConsumer = new ResultConsumerWrapper(method, args, negators, actions, rawConsumer);
							currentResultConsumerWrapper.set(resultConsumer);
							// change detection setup							
							ChangeDetector[] cda=new ChangeDetector[args.length];
							for (int i=0; i<cda.length; i++) {
								cda[i]=newChangeDetector(args[i]);
							}
							
							Object ret=method.invoke(Rule.this, args);
							++invocationCounter;
							
							// change detection
							for (int i=0; i<cda.length; i++) {
								if (cda[i]!=null && cda[i].hasChanged()) {
									update(args[i]);
								}
							}							
								
							if (ret!=null) {
								resultConsumer.consume(ret);
								
								for (int i=0; i<args.length; i++) {									
									Iterator it=negators.iterator();
									while (it.hasNext()) {
										if (Conclusion.object2Negator(args[i], (Negator) it.next())) {
											args[i]=null;
											return i-1; // Last non-negated argument index.
										}
									}									
								}								
							}
							return idx;
						} finally {
							currentResultConsumerWrapper.set(prevResultConsumerWrapper);
						}
					} else if (idx==parameterIndex) {
						return doJoin(args, idx+1, negators, negatorVersions, actions, rawConsumer);						
					} else {
						 // Figure out whether negators shall be applied to collection entries
						boolean applyNegators = negators.getVersion() > negatorVersions[idx];
						negatorVersions[idx]=negators.getVersion();
						
						Iterator it=collections[idx].iterator();
						Z: while (it.hasNext()) {
							args[idx]=it.next();
							
							// Apply negators to the next element if needed
							if (applyNegators) {
								Iterator nit=negators.iterator();
								while (nit.hasNext()) {
									if (Conclusion.object2Negator(args[idx], (Negator) nit.next())) {
										it.remove();
										continue Z;
									}
								}
							} 
							
							int lna = doJoin(args, idx+1, negators, negatorVersions, actions, rawConsumer);
							if (lna<idx) {
								return lna;
							}
							applyNegators = applyNegators || negators.getVersion() > negatorVersions[idx];
						}
						return idx;
					}
				}

				public Class getParameterType() {
					return parameterType;
				}

				public Class[] getFactTypes() {
					Class[] ret = getMethodFactTypes(method.getParameterTypes());
					if (ret!=null) {
						return ret;
					}
					
					Class mrt = method.getReturnType();
					return void.class.equals(mrt) ? new Class[] {} : new Class[] {mrt};
				}
				
			});
		
		}					
	}		
	
	/**
	 * 
	 * @param inferMethodName Methods with this name and one or more arguments are invoked when an object of type 
	 * compatible with one of parameters is posted to the object bus. 
	 * @param removeMethodName Single-argument methods with this name will be invoked when rule set's remove
	 * method with compatible type is invoked. Generally rules shall not implement this method because collection
	 * manager and handle manager take care of removal of the object and conclusions made based on this object from
	 * the knowledge base.
	 * @param acceptMethodName Methods with this name and two arguments - the first of equal type and the second
	 * of <code>AcceptInfo</code> type are used to filter inputs to infer methods with more than one parameter. 
	 * Type of the first argument of <code>accept</code> method and corresponding argument of <code>infer</code>
	 * must be equal. <code>accept()</code> method's return type must be <code>boolean</code>.
	 */
	protected Rule(String inferMethodName, String removeMethodName, String acceptMethodName) {
		this.inferMethodName=inferMethodName;
		this.removeMethodName=removeMethodName;
		this.acceptMethodName=acceptMethodName;
	}

	private long invocationCounter;
	
	/**
	 * @return Number of invocations. Call of reset() method zeroes the counter.
	 */
	public long getInvocationCounter() {
		return invocationCounter;
	}
	
	/**
	 * Wraps result consumer to inject derivations and do other stuff.
	 * @author Pavel
	 *
	 */
	private class ResultConsumerWrapper implements ResultConsumer {

		private Method method;
		private Object[] args;
		private Collection negators;
		private Collection actions;
		private ResultConsumer master;
		
		/**
		 * @param method
		 * @param consumer
		 */
		ResultConsumerWrapper(
				Method method, 
				Object[] args, 
				Collection negators, 
				Collection actions, 
				ResultConsumer resultConsumer) {
			
			this.method = method;
			this.args=args;
			this.negators=negators;
			this.actions=actions;
			this.master = resultConsumer;
		}
		
		public void consume(Object fact) {
			if (fact!=null) {
				if (fact instanceof Conclusion) {
					Derivation derivation=new Derivation(Rule.this, method);
					for (int i=0; i<args.length; i++) {
						derivation.addSourceFact(args[i]);				
					}
					((Conclusion) fact).addDerivation(derivation);				
				}
				
				if (fact instanceof Negator) {
					addNegator((Negator) fact);
				}
				
				if (doTracing) {
					addPostTrace(fact);
				}			
				
				master.consume(fact);
			}
		}
		
		boolean addNegator(Negator negator) {
			if (negators!=null && negators.add(negator)) {
				return true;
			}
			
			return false;
		}

		public void addPostTrace(Object fact) {
			for (int i=0; i<args.length; i++) {
				actions.add(((ActionTracer) owner).createPostAction(args[i], fact));
			}
		}
		
		public void addRemoveTrace(Object fact) {
			for (int i=0; i<args.length; i++) {
				actions.add(((ActionTracer) owner).createPostAction(args[i], fact));
			}
		}		
		
		
	};
	
	
	
	private void createHandlers(final String methodName, Collection handlers) {
		
		Method[] methods=getClass().getMethods();			
		for (int i=0; i<methods.length; i++) {
			if (methods[i].getParameterTypes().length == 1) {
				if (methodName.equals(methods[i].getName())) {
					final Method method=methods[i];
					final Class actualParameterType = method.getParameterTypes()[0];
					
					// Make sure that target method will accept parameterType arguments
					handlers.add(
							new InvocationHandler() {

								public void invoke(Object arg, final ResultConsumer rawConsumer) throws Throwable {																										
									// Invoke only with compatible parameters.
									if (actualParameterType.isInstance(arg)) {
										Object prevResultConsumer=currentResultConsumerWrapper.get();
										try {
																						
											Object[] args = new Object[] {arg};
											Collection actions = doTracing ? new ArrayList() : null;	
											ResultConsumerWrapper resultConsumer = new ResultConsumerWrapper(method, args, null, actions, rawConsumer);
											currentResultConsumerWrapper.set(resultConsumer);
											ChangeDetector cd=newChangeDetector(arg); 
											Object ret=method.invoke(Rule.this, args);
											++invocationCounter;
											if (cd!=null && cd.hasChanged()) {
												update(arg);
											}
											if (ret!=null) {
												resultConsumer.consume(ret);
											}
											
											if (doTracing && !actions.isEmpty()) {
												((ActionTracer) owner).addActions(actions);
											}											
										} finally {
											currentResultConsumerWrapper.set(prevResultConsumer);
										}
									}
								}

								public Class getParameterType() {
									return actualParameterType;
								}
								
								public String toString() {
									return "["+methodName+" handler] Target method: "+method;
								}

								public Class[] getFactTypes() {
									Class[] ret = getMethodFactTypes(method.getParameterTypes());
									if (ret!=null) {
										return ret;
									}
									
									Class mrt = method.getReturnType();
									return void.class.equals(mrt) ? new Class[] {} : new Class[] {mrt};
								}
								
							});
				}
			}
		}		
		
	}
	
	private void createJoinHandlers(String acceptMethodName, String joinMethodName) {
		
		Method[] methods=getClass().getMethods();			
		for (int i=0; i<methods.length; i++) {
			if (methods[i].getParameterTypes().length > 1) {
				if (joinMethodName.equals(methods[i].getName())) {
					Method joinMethod=methods[i];
					Class[] parameterTypes = joinMethod.getParameterTypes();
					Method[] acceptMethods=new Method[parameterTypes.length];					
					for (int j=0; j<parameterTypes.length; j++) { // For each parameter type
						for (int k=0; k<methods.length; k++) { // Find accept method
							Class[] acceptParameterTypes = methods[k].getParameterTypes();
							if (acceptMethodName.equals(methods[k].getName()) 
									&& acceptParameterTypes.length==2
									&& boolean.class.equals(methods[k].getReturnType())
									&& parameterTypes[j].equals(acceptParameterTypes[0])
									&& AcceptInfo.class.equals(acceptParameterTypes[1])) {
								acceptMethods[j]=methods[k];
							}
						}
					}
					
					createJoinHandlers(joinMethod, acceptMethods);
				}					
			}
		}				
	}
	
	private boolean doTracing;
	
	private boolean started;
	
	/**
	 * Locates collection manager.
	 */
	public void start() throws ConfigurationException {
		super.start();
		addHandlers.clear();
		removeHandlers.clear();
		createHandlers(inferMethodName, addHandlers);
		createJoinHandlers(acceptMethodName, inferMethodName);				
		createHandlers(removeMethodName, removeHandlers);
		doTracing = owner instanceof ActionTracer;
		started = true;
	}

	public Collection getInvocationHandlers() {
		return addHandlers;
	}
	
	/**
	 * @return Collection of remove handlers.
	 */
	public Collection getRemoveHandlers() {
		return removeHandlers;
	}
	
	/**
	 * Adds new fact to knowledge base.
	 * Returning value from inference methods has the same effect.
	 * @param fact
	 */
	protected void post(Object fact) {
		if (fact!=null) {
			ResultConsumerWrapper resultConsumer = (ResultConsumerWrapper) currentResultConsumerWrapper.get();			
			if (resultConsumer==null) {
				throw new IllegalStateException("post() can be invoked only within rule method call.");
			}
			
			resultConsumer.consume(fact);
		}
	}	
	
	/**
	 * Invokes remove method of the knowledge base and adds trace action.
	 */
	protected void remove(final Object fact) {
		if (fact!=null) {
			ResultConsumerWrapper resultConsumer = (ResultConsumerWrapper) currentResultConsumerWrapper.get();			
			if (resultConsumer==null) {
				throw new IllegalStateException("remove() can be invoked only within rule method call.");
			}
			
			if (doTracing) {
				resultConsumer.addRemoveTrace(fact);
			}
			
			resultConsumer.consume(new KnowledgeBase.KnowledgeBaseCommand() {

				public void execute(KnowledgeBase knowledgeBase) {
					knowledgeBase.remove(fact);
					
				}
				
			});
		}
	}
	
	/**
	 * Invokes update method of the knowledge base and adds trace action.
	 */
	protected void update(final Object fact) {
		if (fact!=null) {
			ResultConsumerWrapper resultConsumer = (ResultConsumerWrapper) currentResultConsumerWrapper.get();			
			if (resultConsumer==null) {
				throw new IllegalStateException("update() can be invoked only within rule method call.");
			}
			
			if (doTracing) {
				resultConsumer.addRemoveTrace(fact);
			}
			
			resultConsumer.consume(new KnowledgeBase.KnowledgeBaseCommand() {

				public void execute(KnowledgeBase knowledgeBase) {
					knowledgeBase.remove(fact);
					knowledgeBase.add(fact);
				}
				
			});
		}
	}
	
	public void reset() {
		super.reset();
		resetInvocationCounter();
	}
	
	/**
	 * Resets invocation counter.
	 * @return counter value before reset.
	 */
	protected long resetInvocationCounter() {
		long ret = invocationCounter;
		invocationCounter = 0;
		return ret;
	}
	
	// List<Class> -> Class[]
	private Map methodFactTypes = new HashMap();
	
	/**
	 * Rule methods can return facts and post facts. Backward reasoning engine
	 * needs to know types of rule outputs. From rule class introspection
	 * the rule system knows about rule method return type, but it doesn't know
	 * about types posted through <code>post()</code>.
	 * 
	 * Also, return type might not be enough, as returned instances may implement
	 * interfaces which other rules are interested in, but which are not declared
	 * in the rule return type.
	 * 
	 * This method allows rules to inform the inference system about posted types.
	 * <P>
	 * This method shall be invoked before rule is started. Place invocations of 
	 * this method in rule constructors.
	 * <P>
	 * Only rules used in backward reasoning need/must call this method to ensure
	 * proper reasoning. For methods without fact types information set through 
	 * this method, the rule system uses method return type. 
	 * 
	 * @param parameterTypes Reasoning method parameter types. Method name is not 
	 * required as it is known.
	 * @param factTypes Fact types produced by given reasoning (inference) method.
	 * If rule method returns facts as well as posts them, then method return type 
	 * must be included in the factTypes array. 
	 */
	protected final void setMethodFactTypes(Class[] parameterTypes, Class[] factTypes) {
		if (started) {
			throw new IllegalStateException("Method fact types cannot be set once rule is started.");
		}
		
		methodFactTypes.put(new ArrayList(Arrays.asList(parameterTypes)), factTypes);
	}
	
	/**
	 * Convenience method, delegates to <code>setMethodFactTypes(Class[] parameterTypes, Class[] factTypes)</code>
	 * @param parameterType
	 * @param factTypes
	 */
	protected final void setMethodFactTypes(Class parameterType, Class[] factTypes) {
		setMethodFactTypes(new Class[] {parameterType}, factTypes);
	}
	
	/**
	 * Convenience method, delegates to <code>setMethodFactTypes(Class[] parameterTypes, Class[] factTypes)</code>
	 * @param parameterType
	 * @param factTypes
	 */
	protected final void setMethodFactTypes(Class parameterType, Class factType) {
		setMethodFactTypes(parameterType, new Class[] {factType});
	}
	
	/**
	 * Convenience method, delegates to <code>setMethodFactTypes(Class[] parameterTypes, Class[] factTypes)</code>
	 * @param parameterType
	 * @param factTypes
	 */
	protected final void setMethodFactTypes(Class[] parameterTypes, Class factType) {
		setMethodFactTypes(parameterTypes, new Class[] {factType});
	}
	
	private Class[] getMethodFactTypes(Class[] parameterTypes) {
		return (Class[]) methodFactTypes.get(new ArrayList(Arrays.asList(parameterTypes)));
	}
}
