package biz.hammurapi.rules.backwardreasoning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.logging.Logger;

/**
 * Accumulates objects. Allows multiple producers concurrently
 * add objects and multiple consumers iterate over objects.
 * Iterators block until next item is available or all producers
 * are done with producing objects.
 * 
 * Addition of nulls is not allowed.
 * 
 * In some situations last value returned by iterator can be null.
 * Consumers shall check value returned by iterator for null and discard it.
 * 
 * @author Pavel
 *
 */
public class Accumulator {
	
	private static final Logger logger = Logger.getLogger(Accumulator.class.getName());
	
	private Collection producers = new ArrayList();

	private Class type;
	
	/**
	 * Only elements of specified type can be added to accumulator.
	 * @param type
	 */
	public Accumulator(Class type) {
		this.type = type;
	}

	/**
	 * Indicates that one more produces was added to the accumulator.
	 */
	public synchronized void addProducer(Object producer) {
		producers.add(producer);
		logger.fine("Producer "+producer+" added to "+this);
	}
	
	public String toString() {
		return getClass().getName()+" of "+type+", producers: "+producers;
	}
	
	protected interface Link {
		
		Object getValue();
		
		Link peekNext();
		
		void setNext(Link next);
		
		Link getNext() throws InterruptedException;
	}
	
	protected class LinkImpl implements Link {
		
		public LinkImpl(Object value) {
			this.value = value;
		}
		
		Object value;		
		Link next;
		
		/**
		 * Sets next link and notifies waiting iterators.
		 * @param value
		 */
		public synchronized void setNext(Link next) {
			this.next = next;
			notifyAll();			
		}
		
		/**
		 * Retrieves next link. Block until next link is available. 
		 * @return
		 * @throws InterruptedException
		 */
		public synchronized Link getNext() throws InterruptedException {
			while (next==null) {
				onBlock();
				if (next==null) {
					wait();
				}
			}
			return next;
		}

		public Object getValue() {
			return value;
		}

		public Link peekNext() {
			return next;
		}
		
	}
	
	protected final Link END = new Link() {

		public Link getNext() throws InterruptedException {
			throw new UnsupportedOperationException();
		}

		public Object getValue() {
			return null;
		}

		public Link peekNext() {
			throw new UnsupportedOperationException();
		}

		public void setNext(Link next) {
			throw new UnsupportedOperationException();
		}
		
	};
	
	/**
	 * This link 
	 */
	protected final LinkImpl first = new LinkImpl(null);
	
	protected Link current = first;
		
	/**
	 * Returned iterator iterates over available elements and blocks waiting for
	 * more elements from producers or for all producers to finish. remove() is
	 * not supported by returned iterators.
	 * @return Iterator.
	 */
	public Iterator blockingIterator() {
		return new Iterator() {
			
			LinkImpl link = first;

			public boolean hasNext() {
				return link.peekNext() != END;
			}

			public Object next() {
				if (!hasNext()) {
					throw new NoSuchElementException();					
				}
				
				try {				
					Link nxt = link.getNext();
					if (nxt == END) {
						return null;
					}
					link = (LinkImpl) nxt;
					return link.getValue();
				} catch (InterruptedException e) {
					throw new NoSuchElementException("Waiting for the next element was interrupted: "+e);					
				}				
			}

			public void remove() {
				throw new UnsupportedOperationException("remove() is not supported");				
			}
			
		};
	}
	
	public interface AccumulatorIterator extends Iterator {
		
		/**
		 * @return true if no more items are going to be added to the accumulator.
		 */
		boolean isClosed();
	}

	/**
	 * Returned iterator iterates over available elements. It does not block waiting for
	 * more elements from producers or for all producers to finish.
	 * When this iterator reaches the end of elements chain it returns false from hasNext().
	 * If later elements are added, it will return true from hasNext() and can be used
	 * again to navigate further. 
	 * remove() is not supported by returned iterators.
	 * @return Iterator.
	 */
	public AccumulatorIterator iterator() {
		return new AccumulatorIterator() {
			
			LinkImpl link = first;

			public boolean hasNext() {
				return link.peekNext()!=null && link.peekNext()!=END;
			}

			public Object next() {
				if (!hasNext()) {
					throw new NoSuchElementException();					
				}
				
				try {
					link = (LinkImpl) link.getNext();
					return link.getValue();
				} catch (InterruptedException e) {
					throw new NoSuchElementException("Waiting for the next element was interrupted: "+e);					
				}				
			}

			public void remove() {
				throw new UnsupportedOperationException("remove() is not supported");				
			}

			public boolean isClosed() {
				return current.peekNext()==END;
			}
			
		};
	}
	
	/**
	 * Adds object, unblocks waiting iterators.
	 * @param obj
	 */
	public synchronized void add(Object obj) {
		if (obj == null) {
			throw new NullPointerException(this.getClass().getName()+" does not support null elements.");
		}
		if (current.peekNext()==END) {
			throw new IllegalStateException("Accumulator is closed, no more items can be added.");
		}
		if (!type.isInstance(obj)) {
			throw new ClassCastException("Expected "+type.getName()+", got "+obj.getClass().getName());
		}
		logger.finest("----> Element '"+obj+"' added to "+this);
		Link newLink = new LinkImpl(obj);
		newLink.setNext(current.peekNext());
		current.setNext(newLink);
		current = newLink;		
	}
	
	/**
	 * Invocation of this method indicates that no more elements will be added to
	 * accumulator.
	 */
	public synchronized void close(Object producer) {
		logger.fine("Producer "+producer+ " closes "+this);		
		if (!producers.contains(producer)) {
			logger.warning("Producer "+producer+" is not registered with "+this);
		}
		producers.remove(producer);
		if (producers.isEmpty()) {
			current.setNext(END);
			logger.fine("Closed "+this);
		}
	}
	
	/**
	 * This method is invoked before iterator enters blocked state.
	 * Subclasses can override this to provide unblocking logic. 
	 * @param it
	 */
	protected void onBlock() {
		logger.finest("Entering blocked state: "+this);
	}
}
