/*
 * hammurapi-rules @mesopotamia.version@
 * Hammurapi rules engine. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.rules.backwardreasoning;

import java.lang.reflect.UndeclaredThrowableException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.DomConfigurableContainer;
import biz.hammurapi.config.RuntimeConfigurationException;
import biz.hammurapi.dispatch.InvocationHandler;
import biz.hammurapi.dispatch.ResultConsumer;
import biz.hammurapi.rules.CollectionManager;
import biz.hammurapi.rules.Constants;
import biz.hammurapi.rules.FactSource;
import biz.hammurapi.rules.Rule;
import biz.hammurapi.rules.RulesRuntimeException;
import biz.hammurapi.rules.backwardreasoning.Accumulator.AccumulatorIterator;
import biz.hammurapi.util.ExceptionSink;
import biz.hammurapi.util.Worker;
import biz.hammurapi.xml.dom.DOMUtils;

/**
 * Backward reasoning rules container.
 * How to use:
 * <OL>
 * <LI>Instantiate from XML definition.</LI>
 * <LI>Set collection manager if rule set contains join rules (inference methods with more than one parameter).</LI>
 * <LI>Set worker for concurrent reasoning. (Optional)</LI>
 * <LI>Set exception sink. Otherwise all exceptions will be reported to the logger. (Optional)</LI>
 * <LI>Set fact sources.</LI>
 * <LI>Invoke getFacts() for fact/conclusion type of interest.</LI>
 * </OL>
 * 
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class BackwardReasoningRulesContainer extends DomConfigurableContainer implements Constants, FactSource {
	
	private static final Logger logger = Logger.getLogger(BackwardReasoningRulesContainer.class.getName());
	
	// Primarily for troubleshooting.
	private Collection allJobs = new ArrayList();

	private static int[] jobNumberCounter = {0};
	
	/**
	 * Job with number.
	 */
	private abstract class ReasoningJob implements Runnable {
		
		int number;
		int pass;
		String name;
		boolean isComplete;
		boolean isInProgress;
		
		// Each job works on one input fact type.
		protected Class inputType;
		
		public ReasoningJob(String name, Class inputType) {
			synchronized (jobNumberCounter) {
				number = ++jobNumberCounter[0];
			}
			this.name = name;
			this.inputType = inputType;
			synchronized (allJobs) {
				allJobs.add(this);
			}
			logger.fine("New job: "+this);
		}
		
		public String toString() {
			return "Job # "+number+" pass "+pass+" '"+name+"' working on "+inputType.getName()+" ("+getClass().getName()+")";
		}				
		
	}

	/**
	 * Job which processes data from accumulator iterator, i.e. 
	 * new items may become available in the iterator and the job
	 * needs to be repeated until iterator is closed.
	 * 
	 * The job adds itself to incremental jobs queue if there are 
	 * no more items and iterator hasNext() returned false. Incremental
	 * jobs are accumulated and are sent to execution after some progress
	 * has been made by other jobs.
	 * @author Pavel
	 *
	 */
	private abstract class IncrementalJob extends ReasoningJob {
		
		AccumulatorIterator iterator;

		IncrementalJob(String name, Class inputType, AccumulatorIterator iterator) {
			super(name, inputType);
			this.iterator = iterator;
			this.inputType = inputType;
		}

		public void run() {
			logger.fine("Running "+this.toString()+" in "+Thread.currentThread().getName());
			isInProgress = true;
			while (iterator.hasNext()) {
				doIteration(iterator.next());
			}
			
			if (iterator.isClosed()) {
				isComplete = true;
				close();
				synchronized (allJobs) {
					allJobs.remove(this);
					logger.fine("Job finished: "+this.toString()+" in "+Thread.currentThread().getName());
				}
			} else {
				logger.fine("Job sent to next pass: "+this.toString()+" in "+Thread.currentThread().getName());				
				queueIncrementalJob(this);
			}
		}

		/**
		 * This method is invoked for each iterator value.
		 * If the job produced conclusions from that value,
		 * it should invoke onProgress(jobNumber, factTypes) to inform
		 * other jobs that there is additional information 
		 * for them.
		 * @param next
		 */
		abstract void doIteration(Object next);
		
		abstract void close();
		
	}
	
	/**
	 * This class pumps data from one accumulator to another (from subtype to type).
	 * @author Pavel
	 *
	 */
	private class AccumulatorJob extends IncrementalJob {
		
		Accumulator accumulator;
		Class outputType;

		AccumulatorJob(String name, Class inputType, Class outputType, AccumulatorIterator iterator, Accumulator accumulator) {
			super(name, inputType, iterator);
			this.accumulator = accumulator;
			accumulator.addProducer(this);
			this.outputType = outputType;
		}

		void close() {
			logger.fine("Accumulator job closed: "+this);
			accumulator.close(this);
		}

		void doIteration(Object next) {
			if (next!=null) {
				accumulator.add(next);
				onProgress(this, outputType);
			}
		}
		
	}

	/**
	 * Reads facts from external fact source to accumulator.
	 * @author Pavel
	 */
	private class FactSourceJob extends ReasoningJob {

		private FactSource factSource;
		private Accumulator accumulator;

		public FactSourceJob(FactSource factSource, Class inputType) {
			super("Reads "+inputType.getName()+" external fact source "+factSource, inputType);
			this.factSource = factSource;
			TypeEntry te = getTypeEntry(inputType);
			accumulator = te.addJob(this);
		}

		public void run() {
			isInProgress = true;
			try {
				Iterator it = factSource.getFacts(inputType);
				while (it.hasNext()) {
					Object next = it.next();
					if (next!=null) {
						accumulator.add(next);
						onProgress(this, inputType);
					}
				}
			} finally {
				accumulator.close(this);
				isComplete = true;
				synchronized (allJobs) {
					allJobs.remove(this);
				}
			}
		}		
	}		
	
	/**
	 * Executes invocation handler
	 * @author Pavel
	 *
	 */
	private class InvocationHandlerJob extends IncrementalJob {
		
		InvocationHandler invocationHandler;
		Class[] outputTypes;
		ResultConsumer resultConsumer;
		
		// Class -> Accumulator
		Map accumulatorMap = new HashMap();

		InvocationHandlerJob(String name, InvocationHandler invocationHandler) {			
			super(name, invocationHandler.getParameterType(), getTypeEntry(invocationHandler.getParameterType()).lazyIterator());
			this.invocationHandler = invocationHandler;
			
			outputTypes = invocationHandler.getFactTypes();
			for (int i=0; i<outputTypes.length; ++i) {
				accumulatorMap.put(outputTypes[i], getTypeEntry(outputTypes[i]).addJob(this));
			}
			
			resultConsumer = new ResultConsumer() {

				public void consume(Object result) {					
					for (int i=0; i<outputTypes.length; ++i) {
						if (outputTypes[i].isInstance(result)) {
							((Accumulator) accumulatorMap.get(outputTypes[i])).add(result);
							onProgress(InvocationHandlerJob.this, outputTypes[i]);
						}
					}
				}
				
			};
		}

		void close() {
			logger.fine("Invocation handler job closed: "+this);
			for (int i=0; i<outputTypes.length; ++i) {
				((Accumulator) accumulatorMap.get(outputTypes[i])).close(this);
			}
		}

		void doIteration(Object next) {
			if (inputType.isInstance(next)) {
				try {
					invocationHandler.invoke(next, resultConsumer);
				} catch (Exception e) {
					if (exceptionSink==null) {
						logger.log(Level.SEVERE, "Exception in "+this+": "+e, e);
					} else {
						exceptionSink.consume(this, e);
					}
				} catch (Throwable th) {
					if (th instanceof Error) {
						throw (Error) th;
					}
					throw new UndeclaredThrowableException(th);
				}
			}
		}		
	}
	
	// Class -> TypeEntry
	private Map typeEntries = new HashMap();
	
	/**
	 * 
	 * @author Pavel
	 *
	 */
	private class TypeEntry {
		Class type;
		Accumulator theAccumulator; 
		Collection thePendingJobs = new ArrayList();
		
		TypeEntry(Class type) {
			this.type = type;
			theAccumulator = createAccumulator(type);
		}
		
		Accumulator addJob(ReasoningJob job) {
			thePendingJobs.add(job);
			theAccumulator.addProducer(job);
			return theAccumulator;
		}
		
		/**
		 * If there are pending jobs for the accumulator, they are posted to execution.
		 * @return Accumulator with all jobs scheduled for execution.
		 */
		Accumulator getAccumulator() {
			Collection toPost = new ArrayList();
			synchronized (thePendingJobs) {
				Iterator it = thePendingJobs.iterator();
				while (it.hasNext()) {
					ReasoningJob pendingJob = (ReasoningJob) it.next();
					synchronized (pendingJob) {
						if (pendingJob.isComplete || pendingJob.isInProgress) {
							it.remove();
						} else {
							pendingJob.isInProgress = true;
							toPost.add(pendingJob);
						}
					}
				}
			}

			Iterator it = toPost.iterator();
			while (it.hasNext()) {
				postJob((ReasoningJob) it.next());
			}
			return theAccumulator;
		}
		
		
		/**
		 * @return Iterator which activates pending jobs on first access.
		 */
		AccumulatorIterator lazyIterator() {			
			return new Accumulator.AccumulatorIterator() {
				
				AccumulatorIterator master;
				
				AccumulatorIterator getMaster() {
					if (master==null) {
						master = getAccumulator().iterator();
					}
					return master;
				}

				public boolean isClosed() {
					return getMaster().isClosed();
				}

				public boolean hasNext() {
					return getMaster().hasNext();
				}

				public Object next() {
					return getMaster().next();
				}

				public void remove() {
					getMaster().remove();
				}
				
			};
		}
	}
	
	private Worker worker;
	private Collection factSources = new ArrayList();
	private int maxWorkerJobs;
	private ExceptionSink exceptionSink;
	
	private boolean started;
	
	/**
	 * Sets single fact source. 
	 * Resets reasoning results.
	 * @param factSource
	 */
	public void addFactSource(FactSource factSource) {
		if (started) {
			throw new IllegalStateException("Fact source shall be set before container is started");
		}
		factSources.add(factSource);
	}
	
	/**
	 * Exception sink to consume exceptions during reasoning.
	 * If exception sink is not set, exceptions are sent to logger.
	 * @param exceptionSink
	 */
	public void setExceptionSink(ExceptionSink exceptionSink) {
		this.exceptionSink = exceptionSink;
	}
	
	/**
	 * When concurrent reasoning is used, deadlocks are possible depending on worker implementation/configuration. 
	 * If excessive jobs are queued, a situation is possible when all active type jobs wait for a job in a queue to 
	 * complete and that job never gets to execution.   
	 * 
	 * Setting maxWorkerJobs to value greater than zero will limit number of concurrent jobs posted
	 * to worker. Excessive jobs will be executed in invoking thread, to avoid 
	 * being queued by the worker.
	 * @param maxWorkerJobs
	 */
	public void setMaxWorkerJobs(int maxWorkerJobs) {
		this.maxWorkerJobs = maxWorkerJobs;
	}
	
	/**
	 * Sets multiple fact sources. 
	 * Resets reasoning results.
	 * @param factSources
	 */
	public void addFactSources(Collection factSources) {
		if (started) {
			throw new IllegalStateException("Fact sources shall be set before container is started.");
		}
		this.factSources.addAll(factSources);
	}
	
	private TypeEntry getTypeEntry(Class type) {
		TypeEntry ret = (TypeEntry) typeEntries.get(type);
		if (ret == null) {
			ret = new TypeEntry(type);			
			typeEntries.put(type, ret);
		}
		return ret;
	}
	
	// Maximum number of times incremental job can be invoked.
	protected int maxPass;
	
	public void start() throws ConfigurationException {
		super.start();
		started = true;
		Iterator fit = factSources.iterator();
		while (fit.hasNext()) {
			FactSource fs = (FactSource) fit.next();
			Class[] fst = fs.getFactTypes();
			for (int i=0; i<fst.length; ++i) {
				new FactSourceJob(fs, fst[i]);
			}
		}

		Iterator rit = getComponents().iterator();
		while (rit.hasNext()) {
			Object component = rit.next();
			if (component instanceof Rule) {
				Rule rule = (Rule) component;
				Iterator ihit = rule.getInvocationHandlers().iterator();
				while (ihit.hasNext()) {
					InvocationHandler ih = (InvocationHandler) ihit.next();
					new InvocationHandlerJob(rule.getName(), ih);
				}
			}
		}
		
		Set typeSet = typeEntries.keySet();
		types = new Class[typeSet.size()];
		Iterator tit = typeSet.iterator();
		for (int i=0; tit.hasNext(); ++i) {
			types[i] = (Class) tit.next();
		}

		maxPass = allJobs.size()*allJobs.size()*1000;
		
		Iterator it = allJobs.iterator();
		if (logger.isLoggable(Level.FINEST)) {
			logger.finest("All jobs created");
			while (it.hasNext()) {
				logger.finest("\t"+it.next());
			}
		}
	}
	
	/**
	 * Collection manager is required for join (multi-parameter) rules.
	 * @param collectionManager
	 * @throws ConfigurationException 
	 */
	public void setCollectionManager(CollectionManager collectionManager) throws ConfigurationException {
		addComponent(COLLECTION_MANAGER, collectionManager);
	}
	
	/**
	 * Set worker for concurrent reasoning. E.g. if rules access database or 
	 * network resources for different types of facts, retrieval of these facts
	 * can be done in parallel.
	 * @param worker
	 */
	public void setWorker(Worker worker) {
		this.worker = worker;
	}
	
	public Collection getRules() {
		return Collections.unmodifiableCollection(getComponents());
	}
	
	protected String getComponentName(Node node) {
		try {
			return DOMUtils.getSingleNonBlankElementText((Element) node, "name");
		} catch (Exception e) {
			throw new RuntimeConfigurationException("Cannot read rule name: "+e, e);
		}
	}
	
	/**
	 * This method blocks until all reasoning jobs finish execution.
	 * If worker is not set, this method returns immediately.
	 * @throws InterruptedException 
	 */
	public void join() throws InterruptedException {
		if (worker!=null) {
			synchronized (jobCounter) {
				while (jobCounter[0]>0) {
					jobCounter.wait();
				}
			}
		}
	}
		
	private Collection incrementalJobsQueue = new ArrayList();
	
	private void queueIncrementalJob(IncrementalJob job) {
		synchronized (incrementalJobsQueue) {
			incrementalJobsQueue.add(job);
			incrementalJobsQueue.notifyAll(); // just for the case
		}
	}
	
	/**
	 * Method to inform the reasoning system that job made progress
	 * on producing facts of given type.
	 */
	private void onProgress(ReasoningJob job, Class factType) {
		logger.finest("Job "+job+" made progress on "+factType.getName());
		Collection toActivate = new ArrayList();
		synchronized (incrementalJobsQueue) {
			Iterator it = incrementalJobsQueue.iterator();
			while (it.hasNext()) {
				IncrementalJob iJob = (IncrementalJob) it.next();
				if (iJob.inputType.equals(factType)) {
					toActivate.add(iJob);
					logger.fine("Incremental job is sent to additional pass: "+iJob);
					it.remove();
				}
			}
		}

		Iterator it = toActivate.iterator();
		while (it.hasNext()) {
			postJob((ReasoningJob) it.next());
		}
	}
	
	// Every time job goes to execution it is assigned incremental number.
	private int[] passCounter = {0};
	
	// Counts active jobs.
	private int[] jobCounter = {0};

	/**
	 * Executes job either in the current thread or in a worker thread.
	 * @param job
	 */
	private void postJob(final ReasoningJob job) {
		logger.fine("Job posted: "+job);
		synchronized (passCounter) {
			if (passCounter[0] < ++job.pass) {
				passCounter[0] = job.pass;
			}
		}
		
		if (job.pass>maxPass) {
			throw new RulesRuntimeException("Maximum number of reasoning jobs passes ("+maxPass+") has been exceeded for job "+job+". Probably there is an infinite loop in rules logic.");
		}
		
		Runnable workerJob = new Runnable() {

			public void run() {
				try {
					logger.fine("Running reasoning job: "+job);
					job.run();
				} finally {
					synchronized (jobCounter) {
						int jc = --jobCounter[0];
						logger.fine("Finished running reasoning job, job counter "+jc+", " +job);
						if (jc<=0) {
							logger.fine("All reasoning jobs finished");
							jobCounter.notifyAll();
						}
					}
				}					
			}
			
		};
		
		synchronized (jobCounter) {
			++jobCounter[0];
		}
		
		if (worker==null || (maxWorkerJobs>0 && jobCounter[0]>=maxWorkerJobs)) {
			workerJob.run();
		} else {			
			if (!worker.post(workerJob)) {
				workerJob.run();
			}
		}
	}
	
	/**
	 * This thread is activate when there are no more active jobs.
	 * The thread closes pending jobs to unblock pending iterators.
	 * @author Pavel
	 *
	 */
	private class DepletionListener extends Thread {
		
		public DepletionListener() {
			super("Depletion listener");
		}
		
		public void run() {
			synchronized (jobCounter) {
				while (jobCounter[0]>0) {
					try {
						jobCounter.wait();
					} catch (InterruptedException e) {
						logger.warning("Depletion listener was interrupted");
					}
				}
		
				closeIncrementalJobs();
				
				synchronized (jobCounter) {
					depleteionListener = null;
				}
			}
		}
		
	}
	
	private DepletionListener depleteionListener;
	
	/**
	 * If there is exact match, returns iterator for that match, combines
	 * subclasses otherwise.
	 */
	public Iterator getFacts(final Class factType) {
				
		List sources = new ArrayList();
		Iterator it = typeEntries.values().iterator();
		while (it.hasNext()) {
			TypeEntry te = (TypeEntry) it.next();
			if (factType.equals(te.type)) {
				return te.getAccumulator().blockingIterator();
			}
			sources.add(te);
		}
		
		if (sources.isEmpty()) {
			return new Iterator() {

				public boolean hasNext() {
					return false;
				}

				public Object next() {
					throw new NoSuchElementException();
				}

				public void remove() {
					throw new UnsupportedOperationException();
				}
				
			};
		}
		
		Accumulator ret = createAccumulator(factType);
		it = sources.iterator();
		while (it.hasNext()) {
			TypeEntry te = (TypeEntry) it.next();
			postJob(new AccumulatorJob("Superclass combiner", te.type, factType, te.lazyIterator(), ret));
		}

		return ret.blockingIterator();
	}
	
	/**
	 * This implementation creates simple accumulator which doesn't eliminate
	 * duplicates. If several rules come to the same conclusion, these 
	 * conclusions will not be merged. Override this method to produce 
	 * accumulator which eliminates duplicates and merges conclusions. 
	 * @param producers
	 * @return
	 */
	protected Accumulator createAccumulator(Class type) {
		return new Accumulator(type) {
			
			protected void onBlock() {
				logger.finest("Accumulator "+this+" enters blocking state, starting anti-blocking measures");
				onAccumulatorBlock();
			}
			
		};
	}
	
	protected void onAccumulatorBlock() {
		synchronized (jobCounter) {
			if (jobCounter[0]==0) {
				closeIncrementalJobs();
			} else {
				if (depleteionListener==null) {
					depleteionListener = new DepletionListener();
					depleteionListener.start();
				}
			}
		}						
	}
	
	private Class[] types = {};
	
	// Class -> List<FactSource|Rule>
//	private Map typeSourcesMap = new HashMap();

	public Class[] getFactTypes() {
		if (started) {
			return types;
		}
		throw new IllegalStateException("Fact types available after container start.");
	}

	private void closeIncrementalJobs() {
		logger.fine("Closing incremental jobs");
		synchronized (incrementalJobsQueue) {
			Iterator it=incrementalJobsQueue.iterator();
			while (it.hasNext()) {
				IncrementalJob job = (IncrementalJob) it.next();
				if (!job.isComplete) {
					logger.fine("Incremental job closed: "+job);
					job.close();
				}
			}
		}
	}
}