package biz.hammurapi.rules.jsr94;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.rules.ConfigurationException;
import javax.rules.RuleRuntime;
import javax.rules.RuleServiceProviderManager;
import javax.rules.admin.RuleAdministrator;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.rules.jsr94.admin.FileRuleAdministrator;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.DomSerializable;

/**
 * This rule service provider reads configuration from XML files.
 * XML definitions are read from rules home directory. By default it is <code>.hammurapi-rules</code> directory
 * in user home directory. This can be changed by setting <code>biz.hammurapi.rules.jsr94.FileRuleServiceProvider:home</code>
 * property to custom rules home directory name.
 * @author Pavel Vlasov
 * @version ${Revision}
 */
public class FileRuleServiceProvider extends javax.rules.RuleServiceProvider {
	
	private static String home;
	
	static {
		try {
			RuleServiceProviderManager.registerRuleServiceProvider("biz.hammurapi.rules.jsr94.FileRuleServiceProvider", biz.hammurapi.rules.jsr94.FileRuleServiceProvider.class);
		} catch (ConfigurationException e) {
			System.err.println("Could not register biz.hammurapi.rules.jsr94.FileRuleServiceProvider");
			e.printStackTrace();
		}
				
		home=System.getProperty(biz.hammurapi.rules.jsr94.FileRuleServiceProvider.class.getName()+":home");
	}
	
	/**
	 * Registration entry
	 * @author Pavel Vlasov
	 * @version ${Revision}
	 */
	public static class Registration implements DomSerializable {
		private String uri;
		
		/**
		 * @return Returns the uri.
		 */
		public String getUri() {
			return uri;
		}
		
		/**
		 * @param uri The uri to set.
		 */
		public void setUri(String uri) {
			this.uri = uri;
		}
		
		private String ref;
		
		/**
		 * @return Returns the ref.
		 */
		public String getRef() {
			return ref;
		}
			
		/**
		 * @param ref The ref to set.
		 */
		public void setRef(String ref) {
			this.ref = ref;
		}
		
		private Node value;
		
		/**
		 * @return Returns the value.
		 */
		public Node getValue() {
			return value;
		}
		
		/**
		 * @param value The value to set.
		 */
		public void setValue(Node value) {
			this.value = value;
		}

		/* (non-Javadoc)
		 * @see biz.hammurapi.xml.dom.DomSerializable#toDom(org.w3c.dom.Element)
		 */
		public void toDom(Element holder) {
			holder.setAttribute("uri", uri);
			if (ref!=null) {
				holder.setAttribute("ref", ref);			
			} else if (value!=null) {
				Element de=AbstractDomObject.addElement(holder, "definition");
				de.appendChild(holder.getOwnerDocument().importNode(value, true));
			}
			
			if (properties!=null) {
				DOMUtils.toDom(properties, "properties", holder);
			}						
		}

		/* (non-Javadoc)
		 * @see biz.hammurapi.config.DomConfigurable#configure(org.w3c.dom.Node, biz.hammurapi.config.Context)
		 */
		public Registration(Node configNode) throws biz.hammurapi.config.ConfigurationException, XPathExpressionException {
			Element configElement = (Element) configNode;
			uri=configElement.getAttribute("uri");
			if (configElement.hasAttribute("ref")) {
				ref=configElement.getAttribute("ref");
			} else {
				value=DOMUtils.selectSingleNode(configNode, "definition/*");
			}			
			
			Node propertiesNode=DOMUtils.selectSingleNode(configNode, "properties");
			if (propertiesNode!=null) {
				properties=(Map) new DomConfigFactory().create(propertiesNode);
			}
		}	
		

		private Map properties;
		
		/**
		 * @return Returns the properties.
		 */
		public Map getProperties() {
			return properties;
		}

		public Registration(String uri, String ref, Node value, Map properties) {
			super();
			// TODO Auto-generated constructor stub
			this.uri = uri;
			this.ref = ref;
			this.value = value;
			
			if (properties!=null) {
				this.properties = new HashMap(properties);
			}
		}
	}
	
	private Map registrations=new HashMap();
	private RuleRuntime runtime;
	private RuleAdministrator administrator;
	private File regFile;
	
	/**
	 * Constructs rule service provider and reads 
	 * rule definitions from XML file <code>registrations.xml</code> in
	 * rules home directory.
	 *
	 */
	public FileRuleServiceProvider() {
		File homeDir;
		
		if (home==null) {
			homeDir=new File(new File(System.getProperty("user.home")), ".hammurapi-rules");
		} else {
			homeDir=new File(home);
		}
		
		if (homeDir.exists()) {
			if (!homeDir.isDirectory()) {
				errorMessage="Invalid home, not a directory - "+homeDir.getAbsolutePath();
			}
		} else if (!homeDir.mkdirs()) {
			errorMessage="Cannot create home directory "+homeDir.getAbsolutePath();
		}
		
		regFile=new File(homeDir, "registrations.xml");		
		try {
			load();
		} catch (Exception e) {
			errorMessage="Could not load registrations: "+e;
		}				
				
		runtime=new FileRuleRuntime(registrations);
		administrator=new FileRuleAdministrator(registrations, regFile);	
	}

	/**
	 * Loads configuration from XML file
	 * @param regFile
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws FactoryConfigurationError
	 * @throws TransformerException 
	 * @throws biz.hammurapi.config.ConfigurationException 
	 * @throws XPathExpressionException 
	 */
	private void load() throws SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, biz.hammurapi.config.ConfigurationException, XPathExpressionException {
		if (regFile.exists()) {
			synchronized (registrations) {
				registrations.clear();
				Document doc=DOMUtils.parse(regFile);
				NodeList nl=DOMUtils.selectNodeList(doc.getDocumentElement(), "/registrations/registration");
				for (int i=0, l=nl.getLength(); i<l; ++i) {					
					Registration r=new Registration(nl.item(i));
					registrations.put(r.getUri(), r);
				}
			}
		}
	}
	
	private String errorMessage;

	public RuleRuntime getRuleRuntime() throws ConfigurationException {
		if (errorMessage!=null) {
			throw new ConfigurationException(errorMessage);
		}
		
		return runtime;
	}

	public RuleAdministrator getRuleAdministrator()	throws ConfigurationException {
		if (errorMessage!=null) {
			throw new ConfigurationException(errorMessage);
		}
		
		return administrator;
	}		
}
