package biz.hammurapi.rules.jsr94.admin;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.rules.RuleServiceProvider;
import javax.rules.RuleServiceProviderManager;
import javax.rules.admin.LocalRuleExecutionSetProvider;
import javax.rules.admin.Rule;
import javax.rules.admin.RuleAdministrator;
import javax.rules.admin.RuleExecutionSet;
import javax.rules.admin.RuleExecutionSetDeregistrationException;
import javax.rules.admin.RuleExecutionSetProvider;
import javax.rules.admin.RuleExecutionSetRegisterException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.rules.jsr94.FileRuleServiceProvider;
import biz.hammurapi.rules.jsr94.FileRuleServiceProvider.Registration;
import biz.hammurapi.util.StreamPumper;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * This rule administrator stores rulesets in an XML file.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class FileRuleAdministrator implements javax.rules.admin.RuleAdministrator {
	
	private Map registrations;
	private File regFile;

	/**
	 * Creates new instance.
	 * @param registrations
	 * @param homeDir
	 */
	public FileRuleAdministrator(Map registrations, File regFile) {
		this.registrations=registrations;
		this.regFile=regFile;
	}
	
	/**
	 * Loads registrations from a file.
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws FactoryConfigurationError
	 * @throws TransformerException
	 * @throws ConfigurationException
	 * @throws XPathExpressionException 
	 */
	private void load() throws SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, ConfigurationException, XPathExpressionException {
		if (regFile.exists()) {
			registrations.clear();
			Document doc=DOMUtils.parse(regFile);
			NodeList nl=DOMUtils.selectNodeList(doc.getDocumentElement(), "/registrations/registration");
			for (int i=0, l=nl.getLength(); i<l; ++i) {
				Registration r=new FileRuleServiceProvider.Registration(nl.item(i));
				registrations.put(r.getUri(), r);
			}
		}
	}	
	
	/**
	 * Stores registrations to a file.
	 * @throws ParserConfigurationException
	 * @throws FactoryConfigurationError
	 * @throws TransformerException 
	 * @throws IOException 
	 */
	private void store() throws ParserConfigurationException, FactoryConfigurationError, IOException, TransformerException {
		DOMUtils.serialize(registrations2Dom().getDocumentElement(), regFile);
	}

	/**
	 * @return
	 * @throws ParserConfigurationException
	 * @throws FactoryConfigurationError
	 */
	private Document registrations2Dom() throws ParserConfigurationException, FactoryConfigurationError {
		Document doc=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		Element root = AbstractDomObject.addElement(doc, "registrations");
		Iterator rit=registrations.values().iterator();
		while (rit.hasNext()) {
			((DomSerializable) rit.next()).toDom(AbstractDomObject.addElement(root, "registration"));
		}
		return doc;
	}	

	/**
	 * Creates new RuleExecutionSetProvider. Propeties are passed to the provider.
	 */
	public RuleExecutionSetProvider getRuleExecutionSetProvider(Map properties) throws RemoteException {
		return new biz.hammurapi.rules.jsr94.admin.RuleExecutionSetProvider(properties);
	}

	/**
	 * Creates new LocalRuleExecutionSetProvider. Properties are passed to the provider.
	 */
	public LocalRuleExecutionSetProvider getLocalRuleExecutionSetProvider(Map properties) throws RemoteException {
		return new biz.hammurapi.rules.jsr94.admin.LocalRuleExecutionSetProvider(properties);
	}

	/**
	 * Registers rule execution set.
	 * Rule execution set can be registered by reference and by value.
	 * When registered by reference the set url is stored in the file instead of actual XML definition.
	 * To store rule execution set by reference put {"by-reference", Boolean.TRUE} to the properties map.
	 */
	public void registerRuleExecutionSet(String uri, RuleExecutionSet res, Map properties) throws RuleExecutionSetRegisterException, RemoteException {
		if (uri.startsWith("direct:")) {
			throw new RuleExecutionSetRegisterException("Invalid rule set URI, 'direct:' is a reserved prefix.");
		}
		boolean byRef=properties!=null && Boolean.TRUE.equals(properties.get("by-reference"));
		biz.hammurapi.rules.jsr94.admin.RuleExecutionSet cRes = ((biz.hammurapi.rules.jsr94.admin.RuleExecutionSet) res);
		if (byRef && cRes.isModified()) {
			throw new RuleExecutionSetRegisterException("Cannot register modified rule set by reference");
		}
		
		if (byRef && cRes.getUri()==null) {
			throw new RuleExecutionSetRegisterException("Cannot register rule set by reference - URI is null");
		}
		
		synchronized (registrations) {
			try {
				load();
				if (byRef) {
					registrations.put(uri, new FileRuleServiceProvider.Registration(uri, cRes.getUri(), null, cRes.getProperties()));
				} else {
					registrations.put(uri, new FileRuleServiceProvider.Registration(uri, null, cRes.getHolder(), cRes.getProperties()));					
				}
				store();
			} catch (Exception e) {
				throw new RuleExecutionSetRegisterException("Could not deregister rule set with URI: "+uri+", "+e, e);
			}
		}
	}

	/**
	 * Properties are ignored.
	 */
	public void deregisterRuleExecutionSet(String uri, Map properties) throws RuleExecutionSetDeregistrationException, RemoteException {
		synchronized (registrations) {
			try {
				load();
				registrations.remove(uri);
				store();
			} catch (Exception e) {
				throw new RuleExecutionSetDeregistrationException("Could not deregister rule set with URI: "+uri+", "+e, e);
			}			
		}
	}
	
	public static void main(String[] args) throws Exception {
	    System.out.println("Hammurapi Rules administrator. Copyright (C) 2006 Hammurapi Group");
		String ruleServiceProviderClassName = "biz.hammurapi.rules.jsr94.FileRuleServiceProvider";
		Class.forName(ruleServiceProviderClassName);
		RuleServiceProvider serviceProvider = RuleServiceProviderManager.getRuleServiceProvider(ruleServiceProviderClassName);
		RuleAdministrator administrator = serviceProvider.getRuleAdministrator();
			    
	    if (args.length==0) {
	    	printHelpAndExit();
	    } else if ("register".equals(args[0])) {
		    Options options=new Options();
		    		    
		    Option helpOption=OptionBuilder
		    .withDescription("Print this message")
		    .isRequired(false)
		    .create("h");
		    
		    options.addOption(helpOption);

		    Option fileOption=OptionBuilder
		    .withArgName("file")
		    .hasArg()
		    .withDescription("Ruleset file")
		    .isRequired(false)
		    .create("f");
		    
		    options.addOption(fileOption);
		    
		    
		    Option urlOption=OptionBuilder
		    .withArgName("URL")
		    .hasArg()
		    .withDescription("Ruleset URL")
		    .isRequired(false)
		    .create("u");
		    
		    options.addOption(urlOption);
		    
		    Option refOption=OptionBuilder
		    .withArgName("URL")
		    .hasArg()
		    .withDescription("Ruleset URL to be registered by reference")
		    .isRequired(false)
		    .create("r");
		    
		    options.addOption(refOption);
		    
		    Option propertyOption=OptionBuilder
		    .withArgName("key=value")
		    .hasArg()
		    .withDescription("Rule set property")
		    .isRequired(false)
		    .create("p");
		    
		    options.addOption(propertyOption);
		    
		    Option rulePropertyOption=OptionBuilder
		    .withArgName("rule name>:<key>=<value")
		    .hasArg()
		    .withDescription("Rule property")
		    .isRequired(false)
		    .create("R");
		    
		    options.addOption(rulePropertyOption);
		    
		    Option objectFilterOption=OptionBuilder
		    .withArgName("class name")
		    .hasArg()
		    .withDescription("Default object filter")
		    .isRequired(false)
		    .create("o");
		    
		    options.addOption(objectFilterOption);
		    
		    CommandLineParser parser=new PosixParser();
		    CommandLine line=null;
		    
		    try {
		        line=parser.parse(options, args);
		    } catch (org.apache.commons.cli.ParseException e) {
		        System.err.println(e.getMessage());
		        System.err.flush();
		        printHelpAndExit("register <uri>", options);
			}
		    
		    if (line.hasOption("h")) {
		        printHelpAndExit("register <uri>", options);
		    }
		
		    int no=0;
		    if (line.hasOption('f')) {
		    	++no;
		    }
		    if (line.hasOption('u')) {
		    	++no;
		    }
		    if (line.hasOption('r')) {
		    	++no;
		    }
		    
		    if (no==0) {
		    	System.out.println("One of -f, -u or -r options must be specified");
		    	printHelpAndExit("register <uri>", options);
		    }
		    
		    if (no>1) {
		    	System.out.println("Options -f, -u, and -r are mutually exclusive");
		    	printHelpAndExit("register <uri>", options);
		    }
		    
			String[] largs=line.getArgs();
		    if (largs.length!=2) {
		    	System.out.println("Invalid parameters");
		    	printHelpAndExit("register <uri>", options);
		    }
		    
			RuleExecutionSetProvider resp = administrator.getRuleExecutionSetProvider(null);
			
			Map properties=null;
			if (line.hasOption('p')) {
				properties=new HashMap();
				String[] props=line.getOptionValues('p');
				for (int i=0; i<props.length; i++) {
					int idx=props[i].indexOf('=');
					if (idx==-1 || idx==props[i].length()-1 || idx==0) {
						System.out.println("Invalid property definition: "+props[i]);
				    	printHelpAndExit("register <uri>", options);						
					}
					properties.put(props[i].substring(0, idx), props[i].substring(idx+1));					
				}
			}			
		    
		    if (line.hasOption('f')) {
				RuleExecutionSet res = resp.createRuleExecutionSet(DOMUtils.parse(new File(line.getOptionValue('f'))).getDocumentElement(), properties);
				setObjectFilterAndRuleProperties(line, options, res);
				administrator.registerRuleExecutionSet(largs[1], res, null);
		    }
		    
		    if (line.hasOption('u')) {
				RuleExecutionSet res = resp.createRuleExecutionSet(line.getOptionValue('u'), properties);
				setObjectFilterAndRuleProperties(line, options, res);
				administrator.registerRuleExecutionSet(largs[1], res, null);
		    }
		    
		    if (line.hasOption('r')) {
		    	if (line.hasOption('R')) {
			    	System.out.println("Rule properties cannot be set in rule sets registered by reference");
			    	printHelpAndExit("register <uri>", options);		    		
		    	}
		    	
		    	if (line.hasOption('o')) {
			    	System.out.println("Defalut object filter cannot be set in rule sets registered by reference");
			    	printHelpAndExit("register <uri>", options);		    		
		    	}
		    	
				RuleExecutionSet res = resp.createRuleExecutionSet(line.getOptionValue('r'), properties);
				HashMap regProperties = new HashMap();
				regProperties.put("by-reference", Boolean.TRUE);
				administrator.registerRuleExecutionSet(largs[1], res, regProperties);
		    }		    
		    
	    } else if ("deregister".equals(args[0])) {
		    Options options=new Options();
		    
		    Option helpOption=OptionBuilder
		    .withDescription("Print this message")
		    .isRequired(false)
		    .create("h");
		    
		    options.addOption(helpOption);
		    
		    CommandLineParser parser=new PosixParser();
		    CommandLine line=null;
		    
		    try {
		        line=parser.parse(options, args);
		    } catch (org.apache.commons.cli.ParseException e) {
		        System.err.println(e.getMessage());
		        System.err.flush();
		        printHelpAndExit("deregister <uri>", options);
			}
		    
		    if (line.hasOption("h")) {
		        printHelpAndExit("deregister <uri>", options);
		    }
		
			String[] largs=line.getArgs();
		    if (largs.length!=2) {
		    	System.out.println("Invalid parameters");
		    	printHelpAndExit("deregister <uri>", options);
		    }
		    
			administrator.deregisterRuleExecutionSet(largs[1], null);
	    } else if ("list".equals(args[0])) {
		    Options options=new Options();
		    
		    Option helpOption=OptionBuilder
		    .withDescription("Print this message")
		    .isRequired(false)
		    .create("h");
		    
		    options.addOption(helpOption);

		    Option fileOption=OptionBuilder
		    .withArgName("file")
		    .hasArg()
		    .withDescription("Output file")
		    .isRequired(false)
		    .create("f");
		    
		    options.addOption(fileOption);		    
		    
		    Option xmlOption=OptionBuilder
		    .withDescription("Output in XML format")
		    .isRequired(false)
		    .create("x");
		    
		    options.addOption(xmlOption);
		    
		    
		    Option htmlOption=OptionBuilder
		    .withDescription("Output in HTML format")
		    .isRequired(false)
		    .create("H");
		    
		    options.addOption(htmlOption);
		    
		    CommandLineParser parser=new PosixParser();
		    CommandLine line=null;
		    
		    try {
		        line=parser.parse(options, args);
		    } catch (org.apache.commons.cli.ParseException e) {
		        System.err.println(e.getMessage());
		        System.err.flush();
		        printHelpAndExit("list", options);
			}
		    
		    if (line.hasOption("h")) {
		        printHelpAndExit("list", options);
		    }
		
		    OutputStream out;
		    if (line.hasOption('f')) {
		    	out=new FileOutputStream(new File(line.getOptionValue('f')));		    	
		    } else {
		    	out=System.out;
		    }
		    
		    File regFile=((FileRuleAdministrator) administrator).regFile;
		    
		    if (line.hasOption('x')) {
		    	FileInputStream in=new FileInputStream(regFile);
		    	try {
		    		new StreamPumper(in, out, null, false).run();
		    	} finally {
		    		in.close();
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}
		    } else if (line.hasOption('H')) {
		    	try {
			    	DOMUtils.style(
			    			DOMUtils.parse(regFile), 
			    			out,
			    			FileRuleAdministrator.class.getResourceAsStream("list_html.xsl"),
			    			null);
		    	} finally {
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}
		    } else {
		    	try {
			    	DOMUtils.style(
			    			DOMUtils.parse(regFile), 
			    			out,
			    			FileRuleAdministrator.class.getResourceAsStream("list_text.xsl"),
			    			null);
		    	} finally {
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}		    	
		    }
	    } else if ("dump".equals(args[0])) {
		    Options options=new Options();
		    
		    Option helpOption=OptionBuilder
		    .withDescription("Print this message")
		    .isRequired(false)
		    .create("h");
		    
		    options.addOption(helpOption);

		    Option fileOption=OptionBuilder
		    .withArgName("file")
		    .hasArg()
		    .withDescription("Output file")
		    .isRequired(false)
		    .create("f");
		    
		    options.addOption(fileOption);		    
		    
		    Option xmlOption=OptionBuilder
		    .withDescription("Output in XML format")
		    .isRequired(false)
		    .create("x");
		    
		    options.addOption(xmlOption);
		    
		    
		    Option htmlOption=OptionBuilder
		    .withDescription("Output in HTML format")
		    .isRequired(false)
		    .create("H");
		    
		    options.addOption(htmlOption);
		    
		    CommandLineParser parser=new PosixParser();
		    CommandLine line=null;
		    
		    try {
		        line=parser.parse(options, args);
		    } catch (org.apache.commons.cli.ParseException e) {
		        System.err.println(e.getMessage());
		        System.err.flush();
		        printHelpAndExit("dump <uri>", options);
			}
		    
		    if (line.hasOption("h")) {
		        printHelpAndExit("dump <uri>", options);
		    }
		
			String[] largs=line.getArgs();
		    if (largs.length!=2) {
		    	System.out.println("Invalid parameters");
		    	printHelpAndExit("dump <uri>", options);
		    }
		    
		    OutputStream out;
		    if (line.hasOption('f')) {
		    	out=new FileOutputStream(new File(line.getOptionValue('f')));		    	
		    } else {
		    	out=System.out;
		    }
		    
		    Object registration = ((FileRuleAdministrator) administrator).registrations.get(largs[1]);
		    
		    if (line.hasOption('x')) {
		    	try {
			    	DOMUtils.serialize(registration, "registration", new StreamResult(out));
		    	} finally {
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}
		    } else if (line.hasOption('H')) {
		    	try {
			    	DOMUtils.style(
			    			registration,
			    			"registration",
			    			out,
			    			FileRuleAdministrator.class.getResourceAsStream("dump_html.xsl"),
			    			null);
		    	} finally {
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}
		    } else {
		    	try {
			    	DOMUtils.style(
			    			registration,
			    			"registration",
			    			out,
			    			FileRuleAdministrator.class.getResourceAsStream("dump_text.xsl"),
			    			null);
		    	} finally {
		    		if (line.hasOption('f')) {
		    			out.close();
		    		}
		    	}		    	
		    }
	    } else {
	    	System.out.println("Unrecognized command: "+args[0]);
	    	printHelpAndExit();
	    }	    	
    }
	
	/**
	 * Reads default object filter and rule properties from command line and sets them in RuleExecutionSet.
	 * @param line
	 * @param res
	 */
	private static void setObjectFilterAndRuleProperties(CommandLine line, Options options, RuleExecutionSet res) {
		if (line.hasOption('o')) {
			res.setDefaultObjectFilter(line.getOptionValue('o'));
		}
		
		if (line.hasOption('R')) {
			String[] ruleProperties=line.getOptionValues('R');
			
			Z: for (int i=0; i<ruleProperties.length; i++) {
				int idx=ruleProperties[i].indexOf(":");
				if (idx<=0 || idx==ruleProperties[i].length()-1) {
					System.out.println("Rule property format: <Rule name>:<property name>=<property value>");
			    	printHelpAndExit("register <uri>", options);						
				}
				
				String ruleName=ruleProperties[i].substring(0, idx);
				
				int idx2=ruleProperties[i].indexOf("=", idx);
				if (idx2==-1 || idx2==ruleProperties[i].length()-1) {
					System.out.println("Rule property format: <Rule name>:<property name>=<property value>");
			    	printHelpAndExit("register <uri>", options);						
				}
				
				String propertyName=ruleProperties[i].substring(idx+1, idx2);
				String propertyValue=ruleProperties[i].substring(idx2+1);
				
				Iterator it=res.getRules().iterator();
				while (it.hasNext()) {
					Rule rule=(Rule) it.next();
					if (ruleName.equals(rule.getName())) {
						rule.setProperty(propertyName, propertyValue);
						continue Z;
					}
				}
				
				System.out.println("Rule not found: "+ruleName);
		    	printHelpAndExit("register <uri>", options);										
			}
		}
		
	}

	protected static void printHelpAndExit() {
	    System.out.println("Usage: hradmin command [options]");
	    System.out.println("\tCommands:");
	    System.out.println("\t\tregister <uri> {-f <file> | -u <url> | -r <url>} [-p <name>=<value>] [-R <rule name>:<property name>=<property value>] [-o <object filter>]- Registers rule set.");
	    System.out.println("\t\tderegister <uri> - Unregisters rule set.");
	    System.out.println("\t\tlist [-f <file>] {-x|-H} - Lists names of registered rule sets to console or file.");
	    System.out.println("\t\tdump <uri>  [-f <file>] {-x|-H} - Outputs ruleset definition to console or file.");
	    System.out.println("\n\tExecute 'hradmin command -h' for command-specific options");	    
	    
	    System.exit(1);
	}
	
	protected static void printHelpAndExit(String command, Options options) {
	    HelpFormatter formatter=new HelpFormatter();
	    formatter.printHelp("hradmin "+command+" [options]", options, false);
	    System.exit(1);
	}
	

}
