package biz.hammurapi.rules.jsr94.admin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.rules.admin.Rule;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.RuntimeConfigurationException;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.DOMUtils;


class RuleExecutionSet implements javax.rules.admin.RuleExecutionSet {
	/**
	 * Generated ID. 
	 */
	private static final long serialVersionUID = 6333369438749221375L;
	
	private Element holder;
	
	Element getHolder() {
		return holder;
		// Handle properties?
	}
	
	Map getProperties() {
		return properties;
	}
	
	private String uri;
	
	String getUri() {
		return uri;
	}
	
	private List rules;
	private String description;
	private String name;
	private boolean modified;
	
	boolean isModified() {
		return modified;
	}

	/**
	 * Constructs rule execution set from XML definition.
	 * The definition shall contain nested elements <code>name</code>, <code>description</code>, and <code>rules</code>.
	 * Rules are defined as nested <code>rule</code> elements of <code>rules</code> element.
	 * @param holder
	 * @throws ConfigurationException
	 * @throws XPathExpressionException 
	 * @throws TransformerException
	 */
	public RuleExecutionSet(Element holder, Map properties) throws ConfigurationException, XPathExpressionException {
		this.holder=holder;
		name=DOMUtils.getSingleElementText(holder, "name");
		description=DOMUtils.getSingleElementText(holder, "description");
		List rl=new ArrayList();
		NodeList nl=DOMUtils.selectNodeList(holder, "rules/rule");
		for (int i=0, l=nl.getLength(); i<l; ++i) {
			final Element ruleElement = (Element) nl.item(i);
			final String name = DOMUtils.getSingleNonBlankElementText(ruleElement, "name");
			final String description = DOMUtils.getSingleNonBlankElementText(ruleElement, "description");
			
			Rule rule=new Rule() {

				public String getName() {
					return name;
				}

				public String getDescription() {
					return description;
				}
				
				// property cache.
				Map properties=new HashMap();

				public Object getProperty(Object key) {
					// Properties set at runtime
					if (properties.containsKey(key)) {
						return properties.get(key);
					} else if (key instanceof String) { // Properties defined in XML.
						try {
							return DOMUtils.getSingleElementText(ruleElement, (String) key);
						} catch (Exception e) {
							throw new RuntimeConfigurationException("Could not read property '"+key+"': "+e, e);
						}
					}
					
					return null;
				}

				public void setProperty(Object key, Object value) {
					properties.put(key, value); // cache properties set at runtime.
					
					if (key instanceof String) {
						if ("name".equals(key)) {
							throw new IllegalArgumentException("'name' property corresponds to rule name and is readonly");							
						}
						if ("description".equals(key)) {
							throw new IllegalArgumentException("'description' property corresponds to rule description and is readonly");							
						}
						try {
							modified=true;
							Element propertyElement = (Element) DOMUtils.selectSingleNode(ruleElement, (String) key);
							if (propertyElement!=null) {
								ruleElement.removeChild(propertyElement);
							}
							
							if (value!=null) {
								DOMUtils.toDom(value, AbstractDomObject.addElement(ruleElement, (String) key));
							}
						} catch (Exception e) {
							throw new RuntimeConfigurationException("Could not set property "+key+"="+value+": "+e, e);
						}								
					} else {
						throw new IllegalArgumentException("Properties shall have keys of type String");
					}
				}
				
			};
			
			rl.add(rule);
		}
		
		rules=Collections.unmodifiableList(rl);
		
		if (properties!=null) {
			this.properties.putAll(properties);
		}
	}

	/**
	 * Constructs rules execution set from XML definition.
	 * @param uri URI of the definition. Used to store rule sets by reference.
	 * @param holder Parent XML element.
	 * @throws ConfigurationException
	 * @throws XPathExpressionException 
	 * @throws TransformerException
	 */
	public RuleExecutionSet(String uri, Element holder, Map properties) throws ConfigurationException, XPathExpressionException {
		this(holder, properties);
		this.uri=uri;
	}

	/**
	 * @return Rule set name.
	 */
	public String getName() {		
		return name;
	}

	/**
	 * @return Rule set description.
	 */
	public String getDescription() {
		return description;
	}
	
	private Map properties=new HashMap();

	/**
	 * @return Runtime property. These properties are not stored in the rule set definition.
	 */
	public Object getProperty(Object key) {		
		return properties.get(key);
	}

	/**
	 * Sets runtime property
	 */
	public void setProperty(Object key, Object value) {
		properties.put(key, value);
	}

	/**
	 * Sets default object filter.
	 */
	public void setDefaultObjectFilter(String type) {
		try {
			modified=true;
			Element objectFilterElement = (Element) DOMUtils.selectSingleNode(holder, "object-filter");
			if (objectFilterElement!=null) {
				holder.removeChild(objectFilterElement);
			}
			
			if (type!=null) {
				objectFilterElement = AbstractDomObject.addElement(holder, "object-filter");
				objectFilterElement.setAttribute("type", type);
			}
		} catch (Exception e) {
			throw new RuntimeConfigurationException("Could not set object filter: "+e, e);
		}		
	}

	/**
	 * @return Default object filter class.
	 */
	public String getDefaultObjectFilter() {
		try {
			Element objectFilterElement = (Element) DOMUtils.selectSingleNode(holder, "object-filter");
			return objectFilterElement==null ? null : objectFilterElement.getAttribute("type"); 
		} catch (Exception e) {
			throw new RuntimeConfigurationException("Could not set object filter: "+e, e);
		}
	}

	/**
	 * Returns rules.
	 */
	public List getRules() {
		return rules;
	}
}
